import json
import logging
from datetime import datetime, timezone, tzinfo
from io import StringIO
from typing import TYPE_CHECKING, Any, Iterable, Optional, Sequence, Union, cast

import httpx

if TYPE_CHECKING:
    import pandas as pd

from phoenix.client.__generated__ import v1
from phoenix.client.exceptions import DuplicateSpanInfo, InvalidSpanInfo, SpanCreationError
from phoenix.client.helpers.spans import dataframe_to_spans as _dataframe_to_spans
from phoenix.client.types.spans import SpanQuery
from phoenix.client.utils.id_handling import is_node_id

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT_IN_SECONDS = 5
_LOCAL_TIMEZONE = datetime.now(timezone.utc).astimezone().tzinfo
_MAX_SPAN_IDS_PER_REQUEST = 100


class Spans:
    """
    Provides methods for interacting with span resources.

    Example:
        Non-DataFrame methods:
            >>> from phoenix.client import Client
            >>> client = Client()

            # Get spans as list
            >>> spans = client.spans.get_spans(
            ...     project_identifier="my-project",
            ...     limit=100
            ... )

            # Get span annotations as list
            >>> annotations = client.spans.get_span_annotations(
            ...     span_ids=["span1", "span2"],
            ...     project_identifier="my-project"
            ... )

            # Log spans
            >>> spans = [
            ...     {
            ...         "id": "1",
            ...         "name": "test",
            ...         "context": {"trace_id": "123", "span_id": "456"},
            ...     }
            ... ]
            >>> result = client.spans.log_spans(
            ...     project_identifier="my-project",
            ...     spans=spans
            ... )
            >>> print(f"Queued {result['total_queued']} spans")

        DataFrame methods:
            >>> from phoenix.client.types.spans import SpanQuery

            # Get spans as DataFrame
            >>> query = SpanQuery().select(annotations["relevance"])
            >>> df = client.spans.get_spans_dataframe(query=query)

            # Get span annotations as DataFrame
            >>> annotations_df = client.spans.get_span_annotations_dataframe(
            ...     span_ids=["span1", "span2"],
            ...     project_identifier="my-project"
            ... )

    """

    def __init__(self, client: httpx.Client) -> None:
        self._client = client

    def get_spans_dataframe(
        self,
        *,
        query: Optional[SpanQuery] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        limit: int = 1000,
        root_spans_only: Optional[bool] = None,
        project_identifier: Optional[str] = None,
        project_name: Optional[str] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> "pd.DataFrame":
        """
        Retrieves spans based on the provided filter conditions.

        Args:
            query: A SpanQuery object defining the query criteria.
            start_time: Optional start time for filtering.
            end_time: Optional end time for filtering.
            limit: Maximum number of spans to return.
            root_spans_only: Whether to return only root spans.
            project_name: Optional project name to filter by. Deprecated, use `project_identifier`
                to also specify by the project id.
            project_identifier: Optional project identifier (name or id) to filter by.
            timeout: Optional request timeout in seconds.

        Returns:
            pandas DataFrame

        Raises:
            ImportError: If pandas is not installed
        """
        project_name = project_name
        query = query if query else SpanQuery()
        normalized_start_time = _normalize_datetime(start_time)
        normalized_end_time = _normalize_datetime(end_time)

        request_body = {
            "queries": [query.to_dict()],
            "start_time": _to_iso_format(normalized_start_time),
            "end_time": _to_iso_format(normalized_end_time),
            "limit": limit,
            "root_spans_only": root_spans_only,
        }

        try:
            import pandas as pd

            _ = pd  # Prevent unused symbol error

            if project_identifier and project_name:
                raise ValueError("Provide only one of 'project_identifier' or 'project_name'.")
            elif project_identifier and not project_name:
                if is_node_id(project_identifier, node_type="Project"):
                    project_response = self._client.get(
                        url=f"v1/projects/{project_identifier}",
                        headers={"accept": "application/json"},
                        timeout=timeout,
                    )
                    project_response.raise_for_status()
                    project = project_response.json()
                    project_name = project["data"]["name"]
                else:
                    project_name = project_identifier

            response = self._client.post(
                url="v1/spans",
                headers={"accept": "application/json"},
                params={"project_name": project_name} if project_name else None,
                json=request_body,
                timeout=timeout,
            )
            return _process_span_dataframe(response)
        except ImportError:
            raise ImportError(
                "pandas is required to use get_spans_dataframe. "
                "Install it with 'pip install pandas'"
            )

    def get_span_annotations_dataframe(
        self,
        *,
        spans_dataframe: Optional["pd.DataFrame"] = None,
        span_ids: Optional[Iterable[str]] = None,
        spans: Optional[Iterable[v1.Span]] = None,
        project_identifier: str = "default",
        include_annotation_names: Optional[Sequence[str]] = None,
        exclude_annotation_names: Optional[Sequence[str]] = None,
        limit: int = 1000,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> "pd.DataFrame":
        """
        Fetches span annotations and returns them as a pandas DataFrame.

        Exactly one of *spans_dataframe*, *span_ids*, or *spans* should be provided.

        Args:
            spans_dataframe: A DataFrame (typically returned by `get_spans_dataframe`) with a
                `context.span_id` or `span_id` column.
            span_ids: An iterable of span IDs.
            spans: A list of Span objects (typically returned by `get_spans`).
            project_identifier: The project identifier (name or ID) used in the API path.
            include_annotation_names: Optional list of annotation names to include. If provided,
                only annotations with these names will be returned.
            exclude_annotation_names: Optional list of annotation names to exclude from results.
                Defaults to ["note"] to exclude note annotations, which are reserved for notes
                added via the UI.
            limit: Maximum number of annotations returned per request page.
            timeout: Optional request timeout in seconds.

        Returns:
            A DataFrame where each row corresponds to a single span annotation.

        Raises:
            ValueError: If not exactly one of *spans_dataframe*, *span_ids*, or *spans* is provided,
                or if the `context.span_id` or `span_id` column is missing from *spans_dataframe*.
            ImportError: If pandas is not installed.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        if exclude_annotation_names is None:
            exclude_annotation_names = ["note"]

        try:
            import pandas as pd
        except ImportError:  # pragma: no cover
            raise ImportError(
                "pandas is required to use get_span_annotations_dataframe. "
                "Install it with 'pip install pandas'"
            )

        # Validate input parameters
        provided_params = sum(
            [
                spans_dataframe is not None,
                span_ids is not None,
                spans is not None,
            ]
        )
        if provided_params != 1:
            raise ValueError("Provide exactly one of 'spans_dataframe', 'span_ids', or 'spans'.")

        if spans_dataframe is not None:
            span_ids_raw: list[str] = cast(
                list[str], spans_dataframe["context.span_id"].dropna().tolist()
            )
            span_ids_list = list({*span_ids_raw})
        elif span_ids is not None:
            span_ids_list = list({*span_ids})
        else:
            assert spans is not None
            _span_ids = [
                span["context"]["span_id"]
                for span in spans
                if span.get("context", {}).get("span_id")
            ]
            span_ids_list = list(set(s for s in _span_ids if s))

        if not span_ids_list:
            return pd.DataFrame()

        annotations: list[v1.SpanAnnotation] = []
        path = f"v1/projects/{project_identifier}/span_annotations"

        for i in range(0, len(span_ids_list), _MAX_SPAN_IDS_PER_REQUEST):
            batch_ids = span_ids_list[i : i + _MAX_SPAN_IDS_PER_REQUEST]
            cursor: Optional[str] = None
            while True:
                params: dict[str, Union[int, str, Sequence[str]]] = {
                    "span_ids": batch_ids,
                    "limit": limit,
                }
                if include_annotation_names is not None:
                    params["include_annotation_names"] = include_annotation_names
                if exclude_annotation_names:
                    params["exclude_annotation_names"] = exclude_annotation_names
                if cursor:
                    params["cursor"] = cursor

                response = self._client.get(
                    url=path,
                    params=params,
                    headers={"accept": "application/json"},
                    timeout=timeout,
                )
                response.raise_for_status()
                payload = response.json()
                payload = cast(v1.SpanAnnotationsResponseBody, payload)
                batch = cast(list[v1.SpanAnnotation], payload.get("data", []))
                annotations.extend(batch)
                cursor = payload.get("next_cursor")
                if not cursor:
                    break  # finished paginating this batch

        df = pd.DataFrame(annotations)
        df = _flatten_nested_column(df, "result")
        df.rename(columns={"name": "annotation_name"}, inplace=True)
        if not df.empty:
            df.set_index("span_id", inplace=True)  # type: ignore[unused-ignore]
        return df

    def get_span_annotations(
        self,
        *,
        span_ids: Optional[Iterable[str]] = None,
        spans: Optional[Iterable[v1.Span]] = None,
        project_identifier: str,
        include_annotation_names: Optional[Sequence[str]] = None,
        exclude_annotation_names: Optional[Sequence[str]] = None,
        limit: int = 1000,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> list[v1.SpanAnnotation]:
        """
        Fetches span annotations and returns them as a list of SpanAnnotation objects.

        Exactly one of *span_ids* or *spans* should be provided.

        Args:
            span_ids: An iterable of span IDs.
            spans: A list of Span objects (typically returned by `get_spans`).
            project_identifier: The project identifier (name or ID) used in the API path.
            include_annotation_names: Optional list of annotation names to include. If provided,
                only annotations with these names will be returned.
            exclude_annotation_names: Optional list of annotation names to exclude from results.
                Defaults to ["note"] to exclude note annotations, which are reserved for notes
                added via the UI.
            limit: Maximum number of annotations returned per request page.
            timeout: Optional request timeout in seconds.

        Returns:
            A list of SpanAnnotation objects.

        Raises:
            ValueError: If not exactly one of *span_ids* or *spans* is provided.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        if exclude_annotation_names is None:
            exclude_annotation_names = ["note"]

        # Validate input parameters
        provided_params = sum(
            [
                span_ids is not None,
                spans is not None,
            ]
        )
        if provided_params != 1:
            raise ValueError("Provide exactly one of 'span_ids' or 'spans'.")

        if span_ids is not None:
            span_ids_list = list({*span_ids})
        else:  # spans is not None
            assert spans is not None
            _span_ids = [
                span["context"]["span_id"]
                for span in spans
                if span.get("context", {}).get("span_id")
            ]
            span_ids_list = list(set(s for s in _span_ids if s))

        if not span_ids_list:
            return []

        annotations: list[v1.SpanAnnotation] = []
        path = f"v1/projects/{project_identifier}/span_annotations"

        for i in range(0, len(span_ids_list), _MAX_SPAN_IDS_PER_REQUEST):
            batch_ids = span_ids_list[i : i + _MAX_SPAN_IDS_PER_REQUEST]
            cursor: Optional[str] = None
            while True:
                params: dict[str, Union[int, str, Sequence[str]]] = {
                    "span_ids": batch_ids,
                    "limit": limit,
                }
                if include_annotation_names is not None:
                    params["include_annotation_names"] = include_annotation_names
                if exclude_annotation_names:
                    params["exclude_annotation_names"] = exclude_annotation_names
                if cursor:
                    params["cursor"] = cursor

                response = self._client.get(
                    url=path,
                    params=params,
                    headers={"accept": "application/json"},
                    timeout=timeout,
                )
                response.raise_for_status()
                payload = response.json()
                payload = cast(v1.SpanAnnotationsResponseBody, payload)
                batch = cast(list[v1.SpanAnnotation], payload.get("data", []))
                annotations.extend(batch)
                cursor = payload.get("next_cursor")
                if not cursor:
                    break

        return annotations

    def get_spans(
        self,
        *,
        project_identifier: str,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        limit: int = 100,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> list[v1.Span]:
        """
        Retrieves spans with simple filtering options.

        Args:
            project_identifier: The project identifier (name or ID) used in the API path.
            start_time: Optional start time for filtering (inclusive lower bound).
            end_time: Optional end time for filtering (exclusive upper bound).
            limit: Maximum number of spans to return (default: 100).
            timeout: Optional request timeout in seconds.

        Returns:
            A list of Span objects.

        Raises:
            httpx.HTTPStatusError: If the API returns an error response.
        """
        all_spans: list[v1.Span] = []
        cursor: Optional[str] = None
        page_size = min(100, limit)

        while len(all_spans) < limit:
            remaining = limit - len(all_spans)
            current_page_size = min(page_size, remaining)

            params: dict[str, Union[int, str, Sequence[str]]] = {
                "limit": current_page_size,
            }

            if start_time:
                params["start_time"] = start_time.isoformat()
            if end_time:
                params["end_time"] = end_time.isoformat()
            if cursor:
                params["cursor"] = cursor

            response = self._client.get(
                url=f"v1/projects/{project_identifier}/spans",
                params=params,
                headers={"accept": "application/json"},
                timeout=timeout,
            )
            response.raise_for_status()
            payload = response.json()
            payload = cast(v1.SpansResponseBody, payload)

            spans = payload["data"]
            all_spans.extend(spans)

            cursor = payload.get("next_cursor")
            if not cursor or not spans:
                break

        return all_spans[:limit]

    def log_spans(
        self,
        *,
        project_identifier: str,
        spans: Sequence[v1.Span],
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> v1.CreateSpansResponseBody:
        """
        Logs spans to a project.

        If any spans are invalid or duplicates, no spans will be logged and a
        SpanCreationError will be raised with details about the failed spans.

        Args:
            project_identifier: The project identifier (name or ID) used in the API path.
            spans: A sequence of Span objects to log.
            timeout: Optional request timeout in seconds.

        Returns:
            A CreateSpansResponseBody with statistics about the operation. When successful,
            total_queued will equal total_received.

        Raises:
            SpanCreationError: If any spans failed validation (invalid or duplicates).
            httpx.HTTPStatusError: If the API returns an unexpected error response.
            httpx.TimeoutException: If the request times out.
        """
        request_body = v1.CreateSpansRequestBody(data=list(spans))

        response = self._client.post(
            url=f"v1/projects/{project_identifier}/spans",
            json=request_body,
            headers={"accept": "application/json"},
            timeout=timeout,
        )

        if response.status_code not in (400, 422):
            response.raise_for_status()

        result = _parse_log_spans_response(response, spans)
        return result

    def log_spans_dataframe(
        self,
        *,
        project_identifier: str,
        spans_dataframe: "pd.DataFrame",
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> v1.CreateSpansResponseBody:
        """
        Logs spans to a project from a pandas DataFrame.

        If any spans are invalid or duplicates, no spans will be logged and a
        SpanCreationError will be raised with details about the failed spans.

        Args:
            project_identifier: The project identifier (name or ID) used in the API path.
            spans_dataframe: A pandas DataFrame with a `context.span_id` or `span_id` column.
            timeout: Optional request timeout in seconds.
        """
        spans = _dataframe_to_spans(spans_dataframe)
        return self.log_spans(project_identifier=project_identifier, spans=spans, timeout=timeout)


class AsyncSpans:
    """
    Provides async methods for interacting with span resources.

    Example:
        Non-DataFrame methods:
            >>> from phoenix.client import AsyncClient
            >>> client = AsyncClient()

            # Get spans as list
            >>> spans = await client.spans.get_spans(
            ...     project_identifier="my-project",
            ...     limit=100
            ... )

            # Get span annotations as list
            >>> annotations = await client.spans.get_span_annotations(
            ...     span_ids=["span1", "span2"],
            ...     project_identifier="my-project"
            ... )

            # Log spans
            >>> spans = [
            ...     {
            ...         "id": "1",
            ...         "name": "test",
            ...         "context": {"trace_id": "123", "span_id": "456"},
            ...     }
            ... ]
            >>> result = await client.spans.log_spans(
            ...     project_identifier="my-project",
            ...     spans=spans
            ... )
            >>> print(f"Queued {result['total_queued']} spans")

        DataFrame methods:
            >>> from phoenix.client.types.spans import SpanQuery

            # Get spans as DataFrame
            >>> query = SpanQuery().select(annotations["relevance"])
            >>> df = await client.spans.get_spans_dataframe(query=query)

            # Get span annotations as DataFrame
            >>> annotations_df = await client.spans.get_span_annotations_dataframe(
            ...     span_ids=["span1", "span2"],
            ...     project_identifier="my-project"
            ... )

    """

    def __init__(self, client: httpx.AsyncClient) -> None:
        self._client = client

    async def get_spans_dataframe(
        self,
        *,
        query: Optional[SpanQuery] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        limit: int = 1000,
        root_spans_only: Optional[bool] = None,
        project_name: Optional[str] = None,
        project_identifier: Optional[str] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> "pd.DataFrame":
        """
        Retrieves spans based on the provided filter conditions.

        Args:
            query: A SpanQuery object defining the query criteria.
            start_time: Optional start time for filtering.
            end_time: Optional end time for filtering.
            limit: Maximum number of spans to return.
            root_spans_only: Whether to return only root spans.
            project_name: Optional project name to filter by. Deprecated, use `project_identifier`
                to also specify by the project id.
            project_identifier: Optional project identifier (name or id) to filter by.
            timeout: Optional request timeout in seconds.

        Returns:
            pandas DataFrame

        Raises:
            ImportError: If pandas is not installed
        """
        project_name = project_name
        query = query if query else SpanQuery()
        normalized_start_time = _normalize_datetime(start_time)
        normalized_end_time = _normalize_datetime(end_time)

        request_body = {
            "queries": [query.to_dict()],
            "start_time": _to_iso_format(normalized_start_time),
            "end_time": _to_iso_format(normalized_end_time),
            "limit": limit,
            "root_spans_only": root_spans_only,
        }

        try:
            import pandas as pd

            _ = pd  # Prevent unused symbol error

            if project_identifier and project_name:
                raise ValueError("Provide only one of 'project_identifier' or 'project_name'.")
            elif project_identifier and not project_name:
                if is_node_id(project_identifier, node_type="Project"):
                    project_response = await self._client.get(
                        url=f"v1/projects/{project_identifier}",
                        headers={"accept": "application/json"},
                        timeout=timeout,
                    )
                    project_response.raise_for_status()
                    project = project_response.json()
                    project_name = project["data"]["name"]
                else:
                    project_name = project_identifier

            response = await self._client.post(
                url="v1/spans",
                headers={"accept": "application/json"},
                params={"project_name": project_name} if project_name else None,
                json=request_body,
                timeout=timeout,
            )
            await response.aread()
            return _process_span_dataframe(response)
        except ImportError:
            raise ImportError(
                "pandas is required to use get_spans_dataframe. "
                "Install it with 'pip install pandas'"
            )

    async def get_span_annotations_dataframe(
        self,
        *,
        spans_dataframe: Optional["pd.DataFrame"] = None,
        span_ids: Optional[Iterable[str]] = None,
        spans: Optional[Iterable[v1.Span]] = None,
        project_identifier: str,
        include_annotation_names: Optional[Sequence[str]] = None,
        exclude_annotation_names: Optional[Sequence[str]] = None,
        limit: int = 1000,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> "pd.DataFrame":
        """
        Fetches span annotations and returns them as a pandas DataFrame.

        Exactly one of *spans_dataframe*, *span_ids*, or *spans* should be provided.

        Args:
            spans_dataframe: A DataFrame (typically returned by `get_spans_dataframe`) with a
                `context.span_id` or `span_id` column.
            span_ids: An iterable of span IDs.
            spans: A list of Span objects (typically returned by `get_spans`).
            project_identifier: The project identifier (name or ID) used in the API path.
            include_annotation_names: Optional list of annotation names to include. If provided,
                only annotations with these names will be returned.
            exclude_annotation_names: Optional list of annotation names to exclude from results.
                Defaults to ["note"] to exclude note annotations, which are reserved for notes
                added via the UI.
            limit: Maximum number of annotations returned per request page.
            timeout: Optional request timeout in seconds.

        Returns:
            A DataFrame where each row corresponds to a single span annotation.

        Raises:
            ValueError: If not exactly one of *spans_dataframe*, *span_ids*, or *spans* is provided,
                or if the `context.span_id` or `span_id` column is missing from *spans_dataframe*.
            ImportError: If pandas is not installed.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        if exclude_annotation_names is None:
            exclude_annotation_names = ["note"]

        try:
            import pandas as pd
        except ImportError:  # pragma: no cover
            raise ImportError(
                "pandas is required to use get_span_annotations_dataframe. "
                "Install it with 'pip install pandas'"
            )

        # Validate input parameters
        provided_params = sum(
            [
                spans_dataframe is not None,
                span_ids is not None,
                spans is not None,
            ]
        )
        if provided_params != 1:
            raise ValueError("Provide exactly one of 'spans_dataframe', 'span_ids', or 'spans'.")

        if spans_dataframe is not None:
            span_ids_raw: list[str] = cast(
                list[str], spans_dataframe["context.span_id"].dropna().tolist()
            )
            span_ids_list = list({*span_ids_raw})
        elif span_ids is not None:
            span_ids_list = list({*span_ids})
        else:
            assert spans is not None
            _span_ids = [
                span["context"]["span_id"]
                for span in spans
                if span.get("context", {}).get("span_id")
            ]
            span_ids_list = list(set(s for s in _span_ids if s))

        if not span_ids_list:
            return pd.DataFrame()

        annotations: list[v1.SpanAnnotation] = []
        path = f"v1/projects/{project_identifier}/span_annotations"

        for i in range(0, len(span_ids_list), _MAX_SPAN_IDS_PER_REQUEST):
            batch_ids = span_ids_list[i : i + _MAX_SPAN_IDS_PER_REQUEST]
            cursor: Optional[str] = None
            while True:
                params: dict[str, Union[int, str, Sequence[str]]] = {
                    "span_ids": batch_ids,
                    "limit": limit,
                }
                if include_annotation_names is not None:
                    params["include_annotation_names"] = include_annotation_names
                if exclude_annotation_names:
                    params["exclude_annotation_names"] = exclude_annotation_names
                if cursor:
                    params["cursor"] = cursor

                response = await self._client.get(
                    url=path,
                    params=params,
                    headers={"accept": "application/json"},
                    timeout=timeout,
                )
                response.raise_for_status()
                payload = response.json()
                payload = cast(v1.SpanAnnotationsResponseBody, payload)
                batch = cast(list[v1.SpanAnnotation], payload.get("data", []))
                annotations.extend(batch)
                cursor = payload.get("next_cursor")
                if not cursor:
                    break  # finished paginating this batch

        df = pd.DataFrame(annotations)
        df = _flatten_nested_column(df, "result")
        df.rename(columns={"name": "annotation_name"}, inplace=True)
        if not df.empty:
            df.set_index("span_id", inplace=True)  # type: ignore[unused-ignore]
        return df

    async def get_span_annotations(
        self,
        *,
        span_ids: Optional[Iterable[str]] = None,
        spans: Optional[Iterable[v1.Span]] = None,
        project_identifier: str,
        include_annotation_names: Optional[Sequence[str]] = None,
        exclude_annotation_names: Optional[Sequence[str]] = None,
        limit: int = 1000,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> list[v1.SpanAnnotation]:
        """
        Fetches span annotations and returns them as a list of SpanAnnotation objects.

        Exactly one of *span_ids* or *spans* should be provided.

        Args:
            span_ids: An iterable of span IDs.
            spans: A list of Span objects (typically returned by `get_spans`).
            project_identifier: The project identifier (name or ID) used in the API path.
            include_annotation_names: Optional list of annotation names to include. If provided,
                only annotations with these names will be returned.
            exclude_annotation_names: Optional list of annotation names to exclude from results.
                Defaults to ["note"] to exclude note annotations, which are reserved for notes
                added via the UI.
            limit: Maximum number of annotations returned per request page.
            timeout: Optional request timeout in seconds.

        Returns:
            A list of SpanAnnotation objects.

        Raises:
            ValueError: If not exactly one of *span_ids* or *spans* is provided.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        if exclude_annotation_names is None:
            exclude_annotation_names = ["note"]

        # Validate input parameters
        provided_params = sum(
            [
                span_ids is not None,
                spans is not None,
            ]
        )
        if provided_params != 1:
            raise ValueError("Provide exactly one of 'span_ids' or 'spans'.")

        if span_ids is not None:
            span_ids_list = list({*span_ids})
        else:  # spans is not None
            assert spans is not None
            _span_ids = [
                span["context"]["span_id"]
                for span in spans
                if span.get("context", {}).get("span_id")
            ]
            span_ids_list = list(set(s for s in _span_ids if s))

        if not span_ids_list:
            return []

        annotations: list[v1.SpanAnnotation] = []
        path = f"v1/projects/{project_identifier}/span_annotations"

        for i in range(0, len(span_ids_list), _MAX_SPAN_IDS_PER_REQUEST):
            batch_ids = span_ids_list[i : i + _MAX_SPAN_IDS_PER_REQUEST]
            cursor: Optional[str] = None
            while True:
                params: dict[str, Union[int, str, Sequence[str]]] = {
                    "span_ids": batch_ids,
                    "limit": limit,
                }
                if include_annotation_names is not None:
                    params["include_annotation_names"] = include_annotation_names
                if exclude_annotation_names:
                    params["exclude_annotation_names"] = exclude_annotation_names
                if cursor:
                    params["cursor"] = cursor

                response = await self._client.get(
                    url=path,
                    params=params,
                    headers={"accept": "application/json"},
                    timeout=timeout,
                )
                response.raise_for_status()
                payload = response.json()
                payload = cast(v1.SpanAnnotationsResponseBody, payload)
                batch = cast(list[v1.SpanAnnotation], payload.get("data", []))
                annotations.extend(batch)
                cursor = payload.get("next_cursor")
                if not cursor:
                    break

        return annotations

    async def get_spans(
        self,
        *,
        project_identifier: str,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        limit: int = 100,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> list[v1.Span]:
        """
        Retrieves spans with simple filtering options.

        Args:
            project_identifier: The project identifier (name or ID) used in the API path.
            start_time: Optional start time for filtering (inclusive lower bound).
            end_time: Optional end time for filtering (exclusive upper bound).
            limit: Maximum number of spans to return (default: 100).
            timeout: Optional request timeout in seconds.

        Returns:
            A list of Span objects.

        Raises:
            httpx.HTTPStatusError: If the API returns an error response.
        """
        all_spans: list[v1.Span] = []
        cursor: Optional[str] = None
        page_size = min(100, limit)

        while len(all_spans) < limit:
            remaining = limit - len(all_spans)
            current_page_size = min(page_size, remaining)

            params: dict[str, Union[int, str, Sequence[str]]] = {
                "limit": current_page_size,
            }

            if start_time:
                params["start_time"] = start_time.isoformat()
            if end_time:
                params["end_time"] = end_time.isoformat()
            if cursor:
                params["cursor"] = cursor

            response = await self._client.get(
                url=f"v1/projects/{project_identifier}/spans",
                params=params,
                headers={"accept": "application/json"},
                timeout=timeout,
            )
            response.raise_for_status()
            payload = response.json()
            payload = cast(v1.SpansResponseBody, payload)

            spans = payload["data"]
            all_spans.extend(spans)

            cursor = payload.get("next_cursor")
            if not cursor or not spans:
                break

        return all_spans[:limit]

    async def log_spans(
        self,
        *,
        project_identifier: str,
        spans: Sequence[v1.Span],
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> v1.CreateSpansResponseBody:
        """
        Logs spans to a project.

        If any spans are invalid or duplicates, no spans will be logged and a
        SpanCreationError will be raised with details about the failed spans.

        Args:
            project_identifier: The project identifier (name or ID) used in the API path.
            spans: A sequence of Span objects to log.
            timeout: Optional request timeout in seconds.

        Returns:
            A CreateSpansResponseBody with statistics about the operation. When successful,
            total_queued will equal total_received.

        Raises:
            SpanCreationError: If any spans failed validation (invalid or duplicates).
            httpx.HTTPStatusError: If the API returns an unexpected error response.
            httpx.TimeoutException: If the request times out.
        """
        request_body = v1.CreateSpansRequestBody(data=list(spans))

        response = await self._client.post(
            url=f"v1/projects/{project_identifier}/spans",
            json=request_body,
            headers={"accept": "application/json"},
            timeout=timeout,
        )
        if response.status_code not in (400, 422):
            response.raise_for_status()
        result = _parse_log_spans_response(response, spans)

        return result

    async def log_spans_dataframe(
        self,
        *,
        project_identifier: str,
        spans_dataframe: "pd.DataFrame",
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> v1.CreateSpansResponseBody:
        """
        Logs spans to a project from a pandas DataFrame.

        If any spans are invalid or duplicates, no spans will be logged and a
        SpanCreationError will be raised with details about the failed spans.

        Args:
            project_identifier: The project identifier (name or ID) used in the API path.
            spans_dataframe: A pandas DataFrame with a `context.span_id` or `span_id` column.
            timeout: Optional request timeout in seconds.

        Returns:
            A CreateSpansResponseBody with statistics about the operation. When successful,
            total_queued will equal total_received.
        """

        spans = _dataframe_to_spans(spans_dataframe)
        return await self.log_spans(
            project_identifier=project_identifier, spans=spans, timeout=timeout
        )


def _to_iso_format(value: Optional[datetime]) -> Optional[str]:
    return value.isoformat() if value else None


def _decode_df_from_json_string(obj: str) -> "pd.DataFrame":
    import pandas as pd
    from pandas.io.json._table_schema import parse_table_schema  # type: ignore

    df = cast(pd.DataFrame, parse_table_schema(StringIO(obj).read(), False))
    df.index.names = [x.split("_", 1)[1] or None for x in df.index.names]  # type: ignore
    return df.set_axis([x.split("_", 1)[1] for x in df.columns], axis=1)  # type: ignore[override,unused-ignore]


def _normalize_datetime(
    dt: Optional[datetime],
    tz: Optional[tzinfo] = None,
) -> Optional[datetime]:
    """
    If the input datetime is timezone-naive, it is localized as local timezone
    unless tzinfo is specified.
    """
    if not isinstance(dt, datetime):
        return None
    if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
        dt = dt.replace(tzinfo=tz if tz else _LOCAL_TIMEZONE)
    return dt.astimezone(timezone.utc)


def _process_span_dataframe(response: httpx.Response) -> "pd.DataFrame":
    """Processes the httpx response to extract a pandas DataFrame, handling multipart responses."""
    import pandas as pd

    content_type = response.headers.get("Content-Type")
    dfs: list["pd.DataFrame"] = []
    if isinstance(content_type, str) and "multipart/mixed" in content_type:
        if "boundary=" in content_type:
            boundary_token = content_type.split("boundary=")[1].split(";", 1)[0]
        else:
            raise ValueError(
                "Boundary not found in Content-Type header for multipart/mixed response"
            )
        boundary = f"--{boundary_token}"
        text = response.text
        while boundary in text:
            part, text = text.split(boundary, 1)
            if "Content-Type: application/json" in part:
                json_string = part.split("\r\n\r\n", 1)[1].strip()
                df = _decode_df_from_json_string(json_string)
                dfs.append(df)
    else:
        response.raise_for_status()
        logger.warning("Received non-multipart response when expecting dataframe.")

    if dfs:
        return dfs[0]  # only passing in one query
    else:
        return pd.DataFrame()


def _flatten_nested_column(df: "pd.DataFrame", column_name: str) -> "pd.DataFrame":
    import pandas as pd

    if column_name in df.columns:
        # Flatten the nested dictionary column and prefix each resulting column with
        # the original column name (e.g., "result.label").
        nested_df = pd.json_normalize(df[column_name]).rename(  # type: ignore[arg-type]
            columns=lambda col: f"{column_name}.{col}"
        )
        df = pd.concat([df.drop(columns=[column_name]), nested_df], axis=1)
    return df


def _format_log_spans_error_message(
    *,
    total_invalid: int,
    total_duplicates: int,
    invalid_spans: Sequence[InvalidSpanInfo],
    duplicate_spans: Sequence[DuplicateSpanInfo],
) -> str:
    MAX_ERRORS_TO_SHOW = 5
    error_parts: list[str] = []

    if total_invalid > 0:
        error_parts.append(f"Failed to queue {total_invalid} invalid spans:")
        for invalid_span in invalid_spans[:MAX_ERRORS_TO_SHOW]:
            span_id = invalid_span.get("span_id", "unknown")
            error = invalid_span.get("error", "unknown error")
            error_parts.append(f"  - Span {span_id}: {error}")
        if len(invalid_spans) > MAX_ERRORS_TO_SHOW:
            error_parts.append(
                f"  ... and {len(invalid_spans) - MAX_ERRORS_TO_SHOW} more invalid spans"
            )

    if total_duplicates > 0:
        if error_parts:
            error_parts.append("")  # Add blank line
        error_parts.append(f"Found {total_duplicates} duplicate spans:")
        for dup_span in duplicate_spans[:MAX_ERRORS_TO_SHOW]:
            span_id = dup_span.get("span_id", "unknown")
            error_parts.append(f"  - Span {span_id}")
        if len(duplicate_spans) > MAX_ERRORS_TO_SHOW:
            error_parts.append(
                f"  ... and {len(duplicate_spans) - MAX_ERRORS_TO_SHOW} more duplicates"
            )

    return "\n".join(error_parts)


def _parse_log_spans_response(
    response: httpx.Response,
    spans: Sequence[v1.Span],
) -> v1.CreateSpansResponseBody:
    """Parse the response from log spans request."""
    response_data = response.json()

    if response.status_code == 422 and "detail" in response_data:
        error_response = _parse_log_spans_validation_error(response_data, spans)
        _raise_log_spans_error(error_response)

    if response.status_code == 400:
        if "detail" in response_data:
            detail = response_data["detail"]
            # For 400 errors, the server now returns properly formatted JSON in the detail field
            parsed_detail = json.loads(detail)
            _raise_log_spans_error(parsed_detail)
        elif "error" in response_data:
            # Handle case where error data is returned directly (not wrapped in detail)
            _raise_log_spans_error(response_data)

    return cast(v1.CreateSpansResponseBody, response_data)


def _parse_log_spans_validation_error(
    response_data: dict[str, Any],
    spans: Sequence[v1.Span],
) -> v1.CreateSpansResponseBody:
    """Convert FastAPI validation errors to our expected format."""
    invalid_spans: list[InvalidSpanInfo] = []

    for error in response_data.get("detail", []):
        invalid_span = _extract_invalid_span_from_log_spans_error(error, spans)
        if invalid_span:
            invalid_spans.append(invalid_span)

    return cast(
        v1.CreateSpansResponseBody,
        {
            "total_received": len(spans),
            "total_queued": len(spans) - len(invalid_spans),
            "total_duplicates": 0,
            "total_invalid": len(invalid_spans),
            "duplicate_spans": [],
            "invalid_spans": invalid_spans,
        },
    )


def _extract_invalid_span_from_log_spans_error(
    error: dict[str, Any],
    spans: Sequence[v1.Span],
) -> Optional[InvalidSpanInfo]:
    """Extract invalid span info from a validation error."""
    loc_raw = error.get("loc", [])
    if not isinstance(loc_raw, list):
        return None

    loc: list[Any] = loc_raw  # Type annotation to help pyright
    if not (len(loc) >= 3 and loc[0] == "body" and loc[1] == "data" and isinstance(loc[2], int)):
        return None

    span_index = loc[2]
    if span_index >= len(spans):
        return None

    span: Any = spans[span_index]
    span_dict = cast(dict[str, Any], span)
    return {
        "span_id": span_dict.get("context", {}).get("span_id", "unknown"),
        "trace_id": span_dict.get("context", {}).get("trace_id", "unknown"),
        "error": error.get("msg", "Validation error"),
    }


def _raise_log_spans_error(
    error_data: Union[dict[str, Any], v1.CreateSpansResponseBody],
) -> None:
    """Raise SpanCreationError from error response data."""
    total_received = error_data.get("total_received", 0)
    total_queued = error_data.get("total_queued", 0)
    total_invalid = cast(int, error_data.get("total_invalid", 0))
    total_duplicates = cast(int, error_data.get("total_duplicates", 0))
    invalid_spans = cast(list[InvalidSpanInfo], error_data.get("invalid_spans", []))
    duplicate_spans = cast(list[DuplicateSpanInfo], error_data.get("duplicate_spans", []))

    error_msg = _format_log_spans_error_message(
        total_invalid=total_invalid,
        total_duplicates=total_duplicates,
        invalid_spans=invalid_spans,
        duplicate_spans=duplicate_spans,
    )
    raise SpanCreationError(
        message=error_msg,
        invalid_spans=list(invalid_spans),
        duplicate_spans=list(duplicate_spans),
        total_received=total_received,
        total_queued=total_queued,
        total_invalid=total_invalid,
        total_duplicates=total_duplicates,
    )



from __future__ import annotations

import json
import logging
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Iterator,
    Literal,
    Mapping,
    Optional,
    Sequence,
    TypedDict,
    Union,
    overload,
)

from typing_extensions import Required, TypeAlias, assert_never

from phoenix.client.__generated__ import v1
from phoenix.client.utils.template_formatters import TemplateFormatter, to_formatter

if TYPE_CHECKING:
    from openai._client import OpenAI
    from openai.types.chat import (
        ChatCompletionAssistantMessageParam,
        ChatCompletionContentPartImageParam,
        ChatCompletionContentPartInputAudioParam,
        ChatCompletionContentPartParam,
        ChatCompletionContentPartRefusalParam,
        ChatCompletionContentPartTextParam,
        ChatCompletionDeveloperMessageParam,
        ChatCompletionMessageParam,
        ChatCompletionMessageToolCallParam,
        ChatCompletionNamedToolChoiceParam,
        ChatCompletionReasoningEffort,
        ChatCompletionRole,
        ChatCompletionSystemMessageParam,
        ChatCompletionToolChoiceOptionParam,
        ChatCompletionToolMessageParam,
        ChatCompletionToolParam,
        ChatCompletionUserMessageParam,
    )
    from openai.types.chat.chat_completion_assistant_message_param import ContentArrayOfContentPart
    from openai.types.chat.chat_completion_content_part_param import File
    from openai.types.chat.chat_completion_named_tool_choice_param import Function
    from openai.types.chat.completion_create_params import (
        CompletionCreateParamsBase,
        ResponseFormat,
    )
    from openai.types.shared_params import FunctionDefinition, ResponseFormatJSONSchema
    from openai.types.shared_params.response_format_json_schema import JSONSchema

    def _(obj: v1.PromptVersion) -> None:
        messages, kwargs = to_chat_messages_and_kwargs(obj)
        OpenAI().chat.completions.create(messages=messages, **kwargs)

    def __(obj: CompletionCreateParamsBase) -> None:
        create_prompt_version_from_openai(obj)


class _ToolKwargs(TypedDict, total=False):
    parallel_tool_calls: bool
    tool_choice: ChatCompletionToolChoiceOptionParam
    tools: list[ChatCompletionToolParam]


class _InvocationParameters(TypedDict, total=False):
    frequency_penalty: float
    max_completion_tokens: int
    max_tokens: int
    presence_penalty: float
    reasoning_effort: ChatCompletionReasoningEffort
    seed: int
    stop: list[str]
    temperature: float
    top_logprobs: int
    top_p: float


class OpenAIChatCompletionModelKwargs(
    _InvocationParameters,
    _ToolKwargs,
    TypedDict,
    total=False,
):
    model: Required[str]
    response_format: ResponseFormat


_ContentPart: TypeAlias = Union[
    v1.TextContentPart,
    v1.ToolCallContentPart,
    v1.ToolResultContentPart,
]

__all__ = [
    "create_prompt_version_from_openai",
    "to_chat_messages_and_kwargs",
]

logger = logging.getLogger(__name__)


def create_prompt_version_from_openai(
    obj: CompletionCreateParamsBase,
    /,
    *,
    description: Optional[str] = None,
    template_format: Literal["F_STRING", "MUSTACHE", "NONE"] = "MUSTACHE",
    model_provider: Literal["OPENAI", "AZURE_OPENAI", "DEEPSEEK", "XAI", "OLLAMA"] = "OPENAI",
) -> v1.PromptVersionData:
    messages: list[ChatCompletionMessageParam] = list(obj["messages"])
    template = v1.PromptChatTemplate(
        type="chat",
        messages=[_MessageConversion.from_openai(m) for m in messages],
    )
    invocation_parameters = _InvocationParametersConversion.from_openai(
        obj,
        model_provider=model_provider,
    )
    ans = v1.PromptVersionData(
        model_provider=model_provider,
        model_name=obj["model"],
        template=template,
        template_type="CHAT",
        template_format=template_format,
        invocation_parameters=invocation_parameters,
    )
    tool_kwargs: _ToolKwargs = {}
    if "tools" in obj:
        tool_kwargs["tools"] = list(obj["tools"])
    if "tool_choice" in obj:
        tool_kwargs["tool_choice"] = obj["tool_choice"]
    if "parallel_tool_calls" in obj:
        tool_kwargs["parallel_tool_calls"] = obj["parallel_tool_calls"]
    if (tools := _ToolKwargsConversion.from_openai(tool_kwargs)) is not None:
        ans["tools"] = tools
    if "response_format" in obj:
        ans["response_format"] = _ResponseFormatConversion.from_openai(obj["response_format"])
    if description:
        ans["description"] = description
    return ans


def to_chat_messages_and_kwargs(
    obj: v1.PromptVersionData,
    /,
    *,
    variables: Mapping[str, str] = MappingProxyType({}),
    formatter: Optional[TemplateFormatter] = None,
) -> tuple[list[ChatCompletionMessageParam], OpenAIChatCompletionModelKwargs]:
    return (
        list(_to_chat_completion_messages(obj, variables, formatter)),
        _to_model_kwargs(obj),
    )


def _to_model_kwargs(
    obj: v1.PromptVersionData,
) -> OpenAIChatCompletionModelKwargs:
    invocation_parameters: _InvocationParameters = (
        _InvocationParametersConversion.to_openai(obj["invocation_parameters"])
        if "invocation_parameters" in obj
        else {}
    )
    ans: OpenAIChatCompletionModelKwargs = {
        "model": obj["model_name"],
        **invocation_parameters,  # type: ignore[typeddict-item]
    }
    if "tools" in obj:
        tool_kwargs = _ToolKwargsConversion.to_openai(obj["tools"])
        if "tools" in tool_kwargs:
            ans["tools"] = tool_kwargs["tools"]
            if "tool_choice" in tool_kwargs:
                ans["tool_choice"] = tool_kwargs["tool_choice"]
    if "response_format" in obj:
        response_format = obj["response_format"]
        if response_format["type"] == "json_schema":
            ans["response_format"] = _ResponseFormatConversion.to_openai(response_format)
        elif TYPE_CHECKING:
            assert_never(response_format)
    return ans


class _InvocationParametersConversion:
    @staticmethod
    def to_openai(
        obj: Union[
            v1.PromptOpenAIInvocationParameters,
            v1.PromptAzureOpenAIInvocationParameters,
            v1.PromptAnthropicInvocationParameters,
            v1.PromptGoogleInvocationParameters,
            v1.PromptDeepSeekInvocationParameters,
            v1.PromptXAIInvocationParameters,
            v1.PromptOllamaInvocationParameters,
            v1.PromptAwsInvocationParameters,
        ],
    ) -> _InvocationParameters:
        ans: _InvocationParameters = {}
        if obj["type"] == "openai":
            openai_params: v1.PromptOpenAIInvocationParametersContent
            openai_params = obj["openai"]
            if "max_completion_tokens" in openai_params:
                ans["max_completion_tokens"] = openai_params["max_completion_tokens"]
            if "max_tokens" in openai_params:
                ans["max_tokens"] = openai_params["max_tokens"]
            if "temperature" in openai_params:
                ans["temperature"] = openai_params["temperature"]
            if "top_p" in openai_params:
                ans["top_p"] = openai_params["top_p"]
            if "presence_penalty" in openai_params:
                ans["presence_penalty"] = openai_params["presence_penalty"]
            if "frequency_penalty" in openai_params:
                ans["frequency_penalty"] = openai_params["frequency_penalty"]
            if "seed" in openai_params:
                ans["seed"] = openai_params["seed"]
            if "reasoning_effort" in openai_params:
                ans["reasoning_effort"] = openai_params["reasoning_effort"]
        elif obj["type"] == "azure_openai":
            azure_params: v1.PromptAzureOpenAIInvocationParametersContent
            azure_params = obj["azure_openai"]
            if "max_completion_tokens" in azure_params:
                ans["max_completion_tokens"] = azure_params["max_completion_tokens"]
            if "max_tokens" in azure_params:
                ans["max_tokens"] = azure_params["max_tokens"]
            if "temperature" in azure_params:
                ans["temperature"] = azure_params["temperature"]
            if "top_p" in azure_params:
                ans["top_p"] = azure_params["top_p"]
            if "presence_penalty" in azure_params:
                ans["presence_penalty"] = azure_params["presence_penalty"]
            if "frequency_penalty" in azure_params:
                ans["frequency_penalty"] = azure_params["frequency_penalty"]
            if "seed" in azure_params:
                ans["seed"] = azure_params["seed"]
            if "reasoning_effort" in azure_params:
                ans["reasoning_effort"] = azure_params["reasoning_effort"]
        elif obj["type"] == "deepseek":
            deepseek_params: v1.PromptDeepSeekInvocationParametersContent
            deepseek_params = obj["deepseek"]
            if "max_completion_tokens" in deepseek_params:
                ans["max_completion_tokens"] = deepseek_params["max_completion_tokens"]
            if "max_tokens" in deepseek_params:
                ans["max_tokens"] = deepseek_params["max_tokens"]
            if "temperature" in deepseek_params:
                ans["temperature"] = deepseek_params["temperature"]
            if "top_p" in deepseek_params:
                ans["top_p"] = deepseek_params["top_p"]
            if "presence_penalty" in deepseek_params:
                ans["presence_penalty"] = deepseek_params["presence_penalty"]
            if "frequency_penalty" in deepseek_params:
                ans["frequency_penalty"] = deepseek_params["frequency_penalty"]
            if "seed" in deepseek_params:
                ans["seed"] = deepseek_params["seed"]
            if "reasoning_effort" in deepseek_params:
                ans["reasoning_effort"] = deepseek_params["reasoning_effort"]
        elif obj["type"] == "xai":
            xai_params: v1.PromptXAIInvocationParametersContent
            xai_params = obj["xai"]
            if "max_completion_tokens" in xai_params:
                ans["max_completion_tokens"] = xai_params["max_completion_tokens"]
            if "max_tokens" in xai_params:
                ans["max_tokens"] = xai_params["max_tokens"]
            if "temperature" in xai_params:
                ans["temperature"] = xai_params["temperature"]
            if "top_p" in xai_params:
                ans["top_p"] = xai_params["top_p"]
            if "presence_penalty" in xai_params:
                ans["presence_penalty"] = xai_params["presence_penalty"]
            if "frequency_penalty" in xai_params:
                ans["frequency_penalty"] = xai_params["frequency_penalty"]
            if "seed" in xai_params:
                ans["seed"] = xai_params["seed"]
            if "reasoning_effort" in xai_params:
                ans["reasoning_effort"] = xai_params["reasoning_effort"]
        elif obj["type"] == "ollama":
            ollama_params: v1.PromptOllamaInvocationParametersContent
            ollama_params = obj["ollama"]
            if "max_completion_tokens" in ollama_params:
                ans["max_completion_tokens"] = ollama_params["max_completion_tokens"]
            if "max_tokens" in ollama_params:
                ans["max_tokens"] = ollama_params["max_tokens"]
            if "temperature" in ollama_params:
                ans["temperature"] = ollama_params["temperature"]
            if "top_p" in ollama_params:
                ans["top_p"] = ollama_params["top_p"]
            if "presence_penalty" in ollama_params:
                ans["presence_penalty"] = ollama_params["presence_penalty"]
            if "frequency_penalty" in ollama_params:
                ans["frequency_penalty"] = ollama_params["frequency_penalty"]
            if "seed" in ollama_params:
                ans["seed"] = ollama_params["seed"]
            if "reasoning_effort" in ollama_params:
                ans["reasoning_effort"] = ollama_params["reasoning_effort"]
        elif obj["type"] == "anthropic":
            anthropic_params: v1.PromptAnthropicInvocationParametersContent
            anthropic_params = obj["anthropic"]
            if "max_tokens" in anthropic_params:
                ans["max_completion_tokens"] = anthropic_params["max_tokens"]
            if "temperature" in anthropic_params:
                ans["temperature"] = anthropic_params["temperature"]
            if "top_p" in anthropic_params:
                ans["top_p"] = anthropic_params["top_p"]
            if "stop_sequences" in anthropic_params:
                ans["stop"] = list(anthropic_params["stop_sequences"])
        elif obj["type"] == "aws":
            aws_params: v1.PromptAwsInvocationParametersContent
            aws_params = obj["aws"]
            if "max_tokens" in aws_params:
                ans["max_tokens"] = aws_params["max_tokens"]
            if "temperature" in aws_params:
                ans["temperature"] = aws_params["temperature"]
            if "top_p" in aws_params:
                ans["top_p"] = aws_params["top_p"]
        elif obj["type"] == "google":
            google_params: v1.PromptGoogleInvocationParametersContent
            google_params = obj["google"]
            if "max_output_tokens" in google_params:
                ans["max_completion_tokens"] = google_params["max_output_tokens"]
            if "temperature" in google_params:
                ans["temperature"] = google_params["temperature"]
            if "top_p" in google_params:
                ans["top_p"] = google_params["top_p"]
            if "top_k" in google_params:
                ans["top_logprobs"] = google_params["top_k"]
            if "presence_penalty" in google_params:
                ans["presence_penalty"] = google_params["presence_penalty"]
            if "frequency_penalty" in google_params:
                ans["frequency_penalty"] = google_params["frequency_penalty"]
            if "stop_sequences" in google_params:
                ans["stop"] = list(google_params["stop_sequences"])
        elif TYPE_CHECKING:
            assert_never(obj["type"])
        return ans

    @overload
    @staticmethod
    def from_openai(
        obj: CompletionCreateParamsBase,
        /,
        *,
        model_provider: Literal["OPENAI"] = "OPENAI",
    ) -> v1.PromptOpenAIInvocationParameters: ...

    @overload
    @staticmethod
    def from_openai(
        obj: CompletionCreateParamsBase,
        /,
        *,
        model_provider: Literal["AZURE_OPENAI"],
    ) -> v1.PromptAzureOpenAIInvocationParameters: ...

    @overload
    @staticmethod
    def from_openai(
        obj: CompletionCreateParamsBase,
        /,
        *,
        model_provider: Literal["DEEPSEEK"],
    ) -> v1.PromptDeepSeekInvocationParameters: ...

    @overload
    @staticmethod
    def from_openai(
        obj: CompletionCreateParamsBase,
        /,
        *,
        model_provider: Literal["XAI"],
    ) -> v1.PromptXAIInvocationParameters: ...

    @overload
    @staticmethod
    def from_openai(
        obj: CompletionCreateParamsBase,
        /,
        *,
        model_provider: Literal["OLLAMA"],
    ) -> v1.PromptOllamaInvocationParameters: ...

    @staticmethod
    def from_openai(
        obj: CompletionCreateParamsBase,
        /,
        *,
        model_provider: Literal["OPENAI", "AZURE_OPENAI", "DEEPSEEK", "XAI", "OLLAMA"] = "OPENAI",
    ) -> Union[
        v1.PromptOpenAIInvocationParameters,
        v1.PromptAzureOpenAIInvocationParameters,
        v1.PromptDeepSeekInvocationParameters,
        v1.PromptXAIInvocationParameters,
        v1.PromptOllamaInvocationParameters,
    ]:
        content: Union[
            v1.PromptOpenAIInvocationParametersContent,
            v1.PromptAzureOpenAIInvocationParametersContent,
            v1.PromptDeepSeekInvocationParametersContent,
            v1.PromptXAIInvocationParametersContent,
            v1.PromptOllamaInvocationParametersContent,
        ]
        if model_provider == "OPENAI":
            content = v1.PromptOpenAIInvocationParametersContent()
        elif model_provider == "AZURE_OPENAI":
            content = v1.PromptAzureOpenAIInvocationParametersContent()
        elif model_provider == "DEEPSEEK":
            content = v1.PromptDeepSeekInvocationParametersContent()
        elif model_provider == "XAI":
            content = v1.PromptXAIInvocationParametersContent()
        elif model_provider == "OLLAMA":
            content = v1.PromptOllamaInvocationParametersContent()
        else:
            assert_never(model_provider)
        if "max_completion_tokens" in obj and obj["max_completion_tokens"] is not None:
            content["max_completion_tokens"] = obj["max_completion_tokens"]
        if "max_tokens" in obj and obj["max_tokens"] is not None:
            content["max_tokens"] = obj["max_tokens"]
        if "temperature" in obj and obj["temperature"] is not None:
            content["temperature"] = obj["temperature"]
        if "top_p" in obj and obj["top_p"] is not None:
            content["top_p"] = obj["top_p"]
        if "presence_penalty" in obj and obj["presence_penalty"] is not None:
            content["presence_penalty"] = obj["presence_penalty"]
        if "frequency_penalty" in obj and obj["frequency_penalty"] is not None:
            content["frequency_penalty"] = obj["frequency_penalty"]
        if "seed" in obj and obj["seed"] is not None:
            content["seed"] = obj["seed"]
        if "reasoning_effort" in obj:
            v = obj["reasoning_effort"]
            if v in ("low", "medium", "high"):
                content["reasoning_effort"] = v
        if model_provider == "OPENAI":
            return v1.PromptOpenAIInvocationParameters(
                type="openai",
                openai=content,
            )
        elif model_provider == "AZURE_OPENAI":
            return v1.PromptAzureOpenAIInvocationParameters(
                type="azure_openai",
                azure_openai=content,
            )
        elif model_provider == "DEEPSEEK":
            return v1.PromptDeepSeekInvocationParameters(
                type="deepseek",
                deepseek=content,
            )
        elif model_provider == "XAI":
            return v1.PromptXAIInvocationParameters(
                type="xai",
                xai=content,
            )
        elif model_provider == "OLLAMA":
            return v1.PromptOllamaInvocationParameters(
                type="ollama",
                ollama=content,
            )
        else:
            assert_never(model_provider)


def _to_chat_completion_messages(
    obj: v1.PromptVersionData,
    variables: Mapping[str, str],
    formatter: Optional[TemplateFormatter] = None,
    /,
) -> Iterator[ChatCompletionMessageParam]:
    formatter = formatter or to_formatter(obj)
    assert formatter is not None
    template = obj["template"]
    if template["type"] == "chat":
        for message in template["messages"]:
            yield from _MessageConversion.to_openai(message, variables, formatter)
    elif template["type"] == "string":
        raise NotImplementedError
    else:
        assert_never(template)


class _ToolKwargsConversion:
    @staticmethod
    def to_openai(
        obj: Optional[v1.PromptTools],
    ) -> _ToolKwargs:
        ans: _ToolKwargs = {}
        if not obj:
            return ans
        tools: list[ChatCompletionToolParam] = []
        for tool in obj["tools"]:
            if tool["type"] == "function":
                tools.append(_FunctionToolConversion.to_openai(tool))
        if not tools:
            return ans
        ans["tools"] = tools
        if "tool_choice" in obj:
            tool_choice: ChatCompletionToolChoiceOptionParam = _to_tool_choice(obj["tool_choice"])
            ans["tool_choice"] = tool_choice
        if "disable_parallel_tool_calls" in obj:
            v: bool = obj["disable_parallel_tool_calls"]
            ans["parallel_tool_calls"] = not v
        return ans

    @staticmethod
    def from_openai(
        obj: _ToolKwargs,
    ) -> Optional[v1.PromptTools]:
        if not obj or "tools" not in obj:
            return None
        tools: list[v1.PromptToolFunction] = []
        for tool in obj["tools"]:
            if tool["type"] == "function":
                tools.append(_FunctionToolConversion.from_openai(tool))
        if not tools:
            return None
        ans = v1.PromptTools(type="tools", tools=tools)
        if "tool_choice" in obj:
            tc: ChatCompletionToolChoiceOptionParam = obj["tool_choice"]
            ans["tool_choice"] = _from_tool_choice(tc)
        if "parallel_tool_calls" in obj:
            v: bool = obj["parallel_tool_calls"]
            ans["disable_parallel_tool_calls"] = not v
        return ans


def _to_tool_choice(
    obj: Union[
        v1.PromptToolChoiceNone,
        v1.PromptToolChoiceZeroOrMore,
        v1.PromptToolChoiceOneOrMore,
        v1.PromptToolChoiceSpecificFunctionTool,
    ],
) -> ChatCompletionToolChoiceOptionParam:
    if obj["type"] == "none":
        return "none"
    if obj["type"] == "zero_or_more":
        return "auto"
    if obj["type"] == "one_or_more":
        return "required"
    if obj["type"] == "specific_function":
        choice_tool: ChatCompletionNamedToolChoiceParam = {
            "type": "function",
            "function": {"name": obj["function_name"]},
        }
        return choice_tool
    assert_never(obj["type"])


def _from_tool_choice(
    obj: ChatCompletionToolChoiceOptionParam,
) -> Union[
    v1.PromptToolChoiceNone,
    v1.PromptToolChoiceZeroOrMore,
    v1.PromptToolChoiceOneOrMore,
    v1.PromptToolChoiceSpecificFunctionTool,
]:
    if obj == "none":
        choice_none = v1.PromptToolChoiceNone(type="none")
        return choice_none
    if obj == "auto":
        choice_zero_or_more = v1.PromptToolChoiceZeroOrMore(type="zero_or_more")
        return choice_zero_or_more
    if obj == "required":
        choice_one_or_more = v1.PromptToolChoiceOneOrMore(type="one_or_more")
        return choice_one_or_more
    if obj["type"] == "function":
        function: Function = obj["function"]
        choice_function_tool = v1.PromptToolChoiceSpecificFunctionTool(
            type="specific_function",
            function_name=function["name"],
        )
        return choice_function_tool
    assert_never(obj["type"])


class _FunctionToolConversion:
    @staticmethod
    def to_openai(
        obj: v1.PromptToolFunction,
    ) -> ChatCompletionToolParam:
        definition = obj["function"]
        function: FunctionDefinition = {"name": definition["name"]}
        if "description" in definition and isinstance(definition["description"], str):
            function["description"] = definition["description"]
        if "parameters" in definition and isinstance(definition["parameters"], Mapping):
            function["parameters"] = dict(definition["parameters"])
        if "strict" in definition and isinstance(definition["strict"], bool):
            function["strict"] = definition["strict"]
        ans: ChatCompletionToolParam = {"type": "function", "function": function}
        return ans

    @staticmethod
    def from_openai(
        obj: ChatCompletionToolParam,
    ) -> v1.PromptToolFunction:
        definition: FunctionDefinition = obj["function"]
        name = definition["name"]
        function = v1.PromptToolFunctionDefinition(name=name)
        if "description" in definition and isinstance(definition["description"], str):
            function["description"] = definition["description"]
        if "parameters" in definition and isinstance(definition["parameters"], Mapping):
            function["parameters"] = dict(definition["parameters"])
        if "strict" in definition and isinstance(definition["strict"], bool):
            function["strict"] = definition["strict"]
        return v1.PromptToolFunction(type="function", function=function)


class _ResponseFormatConversion:
    @staticmethod
    def to_openai(
        obj: v1.PromptResponseFormatJSONSchema,
    ) -> ResponseFormat:
        definition: v1.PromptResponseFormatJSONSchemaDefinition = obj["json_schema"]
        json_schema: JSONSchema = {
            "name": definition["name"],
        }
        if "schema" in definition and isinstance(definition["schema"], Mapping):
            json_schema["schema"] = dict(definition["schema"])
        if "description" in definition and isinstance(definition["description"], str):
            json_schema["description"] = str(definition["description"])
        if "strict" in definition and isinstance(definition["strict"], bool):
            json_schema["strict"] = bool(definition["strict"])
        response_format_json_schema: ResponseFormatJSONSchema = {
            "type": "json_schema",
            "json_schema": json_schema,
        }
        return response_format_json_schema

    @staticmethod
    def from_openai(
        obj: ResponseFormat,
    ) -> v1.PromptResponseFormatJSONSchema:
        if obj["type"] == "json_schema":
            definition: JSONSchema = obj["json_schema"]
            json_schema = v1.PromptResponseFormatJSONSchemaDefinition(
                name=definition["name"],
            )
            if "schema" in definition and isinstance(definition["schema"], Mapping):
                json_schema["schema"] = dict(definition["schema"])
            if "description" in definition and isinstance(definition["description"], str):
                json_schema["description"] = str(definition["description"])
            if "strict" in definition and isinstance(definition["strict"], bool):
                json_schema["strict"] = bool(definition["strict"])
            response_format_json_schema = v1.PromptResponseFormatJSONSchema(
                type="json_schema",
                json_schema=json_schema,
            )
            return response_format_json_schema
        if obj["type"] == "text":
            raise NotImplementedError
        if obj["type"] == "json_object":
            raise NotImplementedError
        assert_never(obj)


class _MessageConversion:
    @staticmethod
    def to_openai(
        obj: v1.PromptMessage,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> Iterator[ChatCompletionMessageParam]:
        if obj["role"] == "user":
            yield from _UserMessageConversion.to_openai(obj, variables, formatter)
        elif obj["role"] == "system":
            yield from _SystemMessageConversion.to_openai(obj, variables, formatter)
        elif obj["role"] == "developer":
            yield from _DeveloperMessageConversion.to_openai(obj, variables, formatter)
        elif obj["role"] == "assistant" or obj["role"] == "model" or obj["role"] == "ai":
            yield from _AssistantMessageConversion.to_openai(obj, variables, formatter)
        elif obj["role"] == "tool":
            yield from _ToolMessageConversion.to_openai(obj, variables, formatter)
        elif TYPE_CHECKING:
            assert_never(obj["role"])
        else:
            content = list(_ContentPartsConversion.to_openai(obj["content"], variables, formatter))
            yield {"role": obj["role"], "content": content}

    @staticmethod
    def from_openai(
        obj: ChatCompletionMessageParam,
    ) -> v1.PromptMessage:
        if obj["role"] == "user":
            return _UserMessageConversion.from_openai(obj)
        if obj["role"] == "system":
            return _SystemMessageConversion.from_openai(obj)
        if obj["role"] == "developer":
            return _DeveloperMessageConversion.from_openai(obj)
        if obj["role"] == "assistant":
            return _AssistantMessageConversion.from_openai(obj)
        if obj["role"] == "tool":
            return _ToolMessageConversion.from_openai(obj)
        if obj["role"] == "function":
            raise NotImplementedError
        if TYPE_CHECKING:
            assert_never(obj["role"])
        content = list(_ContentPartsConversion.from_openai(obj["content"]))
        return v1.PromptMessage(role=obj["role"], content=content)


class _UserMessageConversion:
    @staticmethod
    def to_openai(
        obj: v1.PromptMessage,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> Iterator[ChatCompletionUserMessageParam]:
        if isinstance(obj["content"], str):
            yield {
                "role": "user",
                "content": formatter.format(obj["content"], variables=variables),
            }
            return
        content = list(_ContentPartsConversion.to_openai(obj["content"], variables, formatter))
        yield _user_msg(content)

    @staticmethod
    def from_openai(
        obj: ChatCompletionUserMessageParam,
        /,
        *,
        role: Literal["user"] = "user",
    ) -> v1.PromptMessage:
        if isinstance(obj["content"], str):
            return v1.PromptMessage(role=role, content=obj["content"])
        content = list(_ContentPartsConversion.from_openai(obj["content"]))
        return v1.PromptMessage(role=role, content=content)


class _SystemMessageConversion:
    @staticmethod
    def to_openai(
        obj: v1.PromptMessage,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> Iterator[ChatCompletionSystemMessageParam]:
        if isinstance(obj["content"], str):
            yield {
                "role": "system",
                "content": formatter.format(obj["content"], variables=variables),
            }
            return
        content = list(
            _ContentPartsConversion.to_openai(obj["content"], variables, formatter, text_only=True)
        )
        if len(content) == 1 and content[0]["type"] == "text":
            yield {
                "role": "system",
                "content": content[0]["text"],
            }
            return
        yield {
            "role": "system",
            "content": content,
        }

    @staticmethod
    def from_openai(
        obj: ChatCompletionSystemMessageParam,
        /,
        *,
        role: Literal["system"] = "system",
    ) -> v1.PromptMessage:
        if isinstance(obj["content"], str):
            return v1.PromptMessage(role=role, content=obj["content"])
        content = list(_ContentPartsConversion.from_openai(obj["content"]))
        return v1.PromptMessage(role=role, content=content)


class _DeveloperMessageConversion:
    @staticmethod
    def to_openai(
        obj: v1.PromptMessage,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> Iterator[ChatCompletionDeveloperMessageParam]:
        if isinstance(obj["content"], str):
            yield {
                "role": "developer",
                "content": formatter.format(obj["content"], variables=variables),
            }
            return
        content = list(
            _ContentPartsConversion.to_openai(obj["content"], variables, formatter, text_only=True)
        )
        if len(content) == 1 and content[0]["type"] == "text":
            yield {
                "role": "developer",
                "content": content[0]["text"],
            }
            return
        yield {
            "role": "developer",
            "content": content,
        }

    @staticmethod
    def from_openai(
        obj: ChatCompletionDeveloperMessageParam,
        /,
        *,
        role: Literal["developer"] = "developer",
    ) -> v1.PromptMessage:
        if isinstance(obj["content"], str):
            return v1.PromptMessage(role=role, content=obj["content"])
        content = list(_ContentPartsConversion.from_openai(obj["content"]))
        return v1.PromptMessage(role=role, content=content)


class _AssistantMessageConversion:
    @staticmethod
    def to_openai(
        obj: v1.PromptMessage,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> Iterator[ChatCompletionAssistantMessageParam]:
        content: list[ContentArrayOfContentPart] = []
        tool_calls: list[ChatCompletionMessageToolCallParam] = []
        if isinstance(obj["content"], str):
            yield {
                "role": "assistant",
                "content": formatter.format(obj["content"], variables=variables),
            }
            return
        for part in obj["content"]:
            if part["type"] == "tool_call":
                tool_calls.append(
                    _ToolCallContentPartConversion.to_openai(part, variables, formatter)
                )
            elif part["type"] == "text":
                content.append(_TextContentPartConversion.to_openai(part, variables, formatter))
            elif part["type"] == "tool_result":
                continue
            elif TYPE_CHECKING:
                assert_never(part)
        if content:
            ans = _assistant_msg(content)
            if tool_calls:
                ans["tool_calls"] = tool_calls
            yield ans
        elif tool_calls:
            yield {"role": "assistant", "tool_calls": tool_calls}

    @staticmethod
    def from_openai(
        obj: ChatCompletionAssistantMessageParam,
        /,
        *,
        role: Literal["assistant"] = "assistant",
    ) -> v1.PromptMessage:
        content: list[_ContentPart] = []
        if "content" in obj and obj["content"] is not None:
            content.extend(_ContentPartsConversion.from_openai(obj["content"]))
        if "tool_calls" in obj and (tool_calls := obj["tool_calls"]):
            content.extend(map(_ToolCallContentPartConversion.from_openai, tool_calls))
        if len(content) == 1 and content[0]["type"] == "text":
            return v1.PromptMessage(role=role, content=content[0]["text"])
        return v1.PromptMessage(role=role, content=content)


class _ToolMessageConversion:
    @staticmethod
    def to_openai(
        obj: v1.PromptMessage,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> Iterator[ChatCompletionToolMessageParam]:
        current_tool_call_id: Optional[str] = None
        current_content: list[ChatCompletionContentPartTextParam] = []
        if isinstance(obj["content"], str):
            yield {
                "role": "tool",
                "tool_call_id": "",
                "content": formatter.format(obj["content"], variables=variables),
            }
            return
        for part in obj["content"]:
            if part["type"] == "tool_result":
                tool_call_id = part["tool_call_id"]
                if (
                    current_tool_call_id is not None
                    and current_tool_call_id != tool_call_id
                    and current_content
                ):
                    yield _tool_msg(tool_call_id=current_tool_call_id, content=current_content)
                    current_content = []
                current_tool_call_id = tool_call_id
                current_content.append(
                    {
                        "type": "text",
                        "text": _str_tool_result(part["tool_result"]),
                    }
                )
            elif part["type"] == "text":
                continue
            elif part["type"] == "tool_call":
                continue
            elif TYPE_CHECKING:
                assert_never(part)
        if current_tool_call_id is not None and current_content:
            yield _tool_msg(tool_call_id=current_tool_call_id, content=current_content)

    @staticmethod
    def from_openai(
        obj: ChatCompletionToolMessageParam,
        /,
        *,
        role: Literal["tool"] = "tool",
    ) -> v1.PromptMessage:
        if isinstance(obj["content"], str):
            return v1.PromptMessage(
                role="tool",
                content=[
                    v1.ToolResultContentPart(
                        type="tool_result",
                        tool_call_id=obj["tool_call_id"],
                        tool_result=obj["content"],
                    )
                ],
            )
        content: list[_ContentPart] = []
        for part in obj["content"]:
            if part["type"] == "text":
                content.append(
                    v1.ToolResultContentPart(
                        type="tool_result",
                        tool_call_id=obj["tool_call_id"],
                        tool_result=part["text"],
                    )
                )
            elif TYPE_CHECKING:
                assert_never(part)
        return v1.PromptMessage(role=role, content=content)


def _str_tool_result(
    obj: Any,
) -> str:
    if isinstance(obj, (dict, list)):
        return json.dumps(obj)
    return str(obj)


class _ToolCallContentPartConversion:
    @staticmethod
    def to_openai(
        obj: v1.ToolCallContentPart,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> ChatCompletionMessageToolCallParam:
        id_ = obj["tool_call_id"]
        tool_call = obj["tool_call"]
        name = tool_call["name"]
        arguments = tool_call["arguments"] if "arguments" in tool_call else "{}"
        return {
            "id": id_,
            "function": {
                "name": name,
                "arguments": arguments,
            },
            "type": "function",
        }

    @staticmethod
    def from_openai(
        obj: ChatCompletionMessageToolCallParam,
    ) -> v1.ToolCallContentPart:
        return v1.ToolCallContentPart(
            type="tool_call",
            tool_call_id=obj["id"],
            tool_call=v1.ToolCallFunction(
                type="function",
                name=obj["function"]["name"],
                arguments=obj["function"]["arguments"],
            ),
        )


class _ContentPartsConversion:
    @overload
    @staticmethod
    def to_openai(
        parts: Sequence[_ContentPart],
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
        *,
        text_only: Literal[True] = True,
    ) -> Iterator[ChatCompletionContentPartTextParam]: ...

    @overload
    @staticmethod
    def to_openai(
        parts: Sequence[_ContentPart],
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
        *,
        text_only: bool,
    ) -> Iterator[ChatCompletionContentPartParam]: ...

    @staticmethod
    def to_openai(
        parts: Sequence[_ContentPart],
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
        *,
        text_only: bool = False,
    ) -> Iterator[Any]:
        for part in parts:
            if part["type"] == "text":
                yield _TextContentPartConversion.to_openai(part, variables, formatter)
            elif text_only:
                continue
            elif part["type"] == "tool_call":
                continue
            elif part["type"] == "tool_result":
                continue
            elif TYPE_CHECKING:
                assert_never(part)

    @staticmethod
    def from_openai(
        obj: Union[
            str,
            Iterable[
                Union[
                    ChatCompletionContentPartTextParam,
                    ChatCompletionContentPartImageParam,
                    ChatCompletionContentPartInputAudioParam,
                    ChatCompletionContentPartRefusalParam,
                    File,
                ]
            ],
            None,
        ],
    ) -> Iterator[_ContentPart]:
        if isinstance(obj, str):
            yield v1.TextContentPart(type="text", text=obj)
            return
        for part in obj or ():
            if part["type"] == "text":
                yield _TextContentPartConversion.from_openai(part)
            elif part["type"] == "image_url":
                continue
            elif part["type"] == "input_audio":
                continue
            elif part["type"] == "refusal":
                continue
            elif part["type"] == "file":
                continue
            elif TYPE_CHECKING:
                assert_never(part["type"])


class _TextContentPartConversion:
    @staticmethod
    def to_openai(
        obj: v1.TextContentPart,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> ChatCompletionContentPartTextParam:
        text = formatter.format(obj["text"], variables=variables)
        return {"type": "text", "text": text}

    @staticmethod
    def from_openai(
        obj: ChatCompletionContentPartTextParam,
    ) -> v1.TextContentPart:
        return v1.TextContentPart(type="text", text=obj["text"])


def _user_msg(
    content: Sequence[ChatCompletionContentPartParam],
) -> ChatCompletionUserMessageParam:
    if len(content) == 1 and content[0]["type"] == "text":
        return {
            "role": "user",
            "content": content[0]["text"],
        }
    return {
        "role": "user",
        "content": content,
    }


def _assistant_msg(
    content: Sequence[ContentArrayOfContentPart],
) -> ChatCompletionAssistantMessageParam:
    if len(content) == 1 and content[0]["type"] == "text":
        return {
            "role": "assistant",
            "content": content[0]["text"],
        }
    return {
        "role": "assistant",
        "content": content,
    }


def _tool_msg(
    tool_call_id: str,
    content: Sequence[ChatCompletionContentPartTextParam],
) -> ChatCompletionToolMessageParam:
    if len(content) == 1 and content[0]["type"] == "text":
        return {
            "role": "tool",
            "tool_call_id": tool_call_id,
            "content": content[0]["text"],
        }
    return {
        "role": "tool",
        "tool_call_id": tool_call_id,
        "content": content,
    }


class _RoleConversion:
    @staticmethod
    def to_openai(
        obj: v1.PromptMessage,
    ) -> ChatCompletionRole:
        role = obj["role"]
        if role == "user":
            return "user"
        if role == "assistant":
            return "assistant"
        if role == "model":
            return "assistant"
        if role == "ai":
            return "assistant"
        if role == "system":
            return "system"
        if role == "developer":
            return "developer"
        if role == "tool":
            return "tool"
        if TYPE_CHECKING:
            assert_never(role)
        return role

    @staticmethod
    def from_openai(
        obj: ChatCompletionMessageParam,
    ) -> Literal["user", "assistant", "tool", "system", "developer"]:
        if obj["role"] == "user":
            return "user"
        if obj["role"] == "system":
            return "system"
        if obj["role"] == "developer":
            return "developer"
        if obj["role"] == "assistant":
            return "assistant"
        if obj["role"] == "tool":
            return "tool"
        if obj["role"] == "function":
            return "tool"
        if TYPE_CHECKING:
            assert_never(obj["role"])
        return obj["role"]



from __future__ import annotations

import logging
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Iterator,
    Literal,
    Mapping,
    Optional,
    Sequence,
    TypedDict,
    Union,
    cast,
)

from typing_extensions import Required, TypeAlias, assert_never

from phoenix.client.__generated__ import v1
from phoenix.client.utils.template_formatters import TemplateFormatter, to_formatter

if TYPE_CHECKING:
    from google.generativeai import protos
    from google.generativeai.generative_models import GenerativeModel
    from google.generativeai.types import GenerationConfig, content_types

    _ContentPart: TypeAlias = Union[
        v1.TextContentPart,
        v1.ToolCallContentPart,
        v1.ToolResultContentPart,
    ]

    def _(obj: v1.PromptVersionData) -> None:
        messages, kwargs = to_chat_messages_and_kwargs(obj)
        GenerativeModel(**kwargs)
        _: Iterable[protos.Content] = messages


class _ToolKwargs(TypedDict, total=False):
    tool_config: protos.ToolConfig
    tools: list[content_types.Tool]


class GoogleModelKwargs(_ToolKwargs, TypedDict, total=False):
    model_name: Required[str]
    generation_config: GenerationConfig
    system_instruction: str | list[str]


logger = logging.getLogger(__name__)

__all__ = [
    "create_prompt_version_from_google",
    "to_chat_messages_and_kwargs",
]


def create_prompt_version_from_google(
    obj: Any,
    /,
    *,
    description: Optional[str] = None,
    template_format: Literal["F_STRING", "MUSTACHE", "NONE"] = "MUSTACHE",
    model_provider: Literal["GOOGLE"] = "GOOGLE",
) -> v1.PromptVersionData:
    raise NotImplementedError


def to_chat_messages_and_kwargs(
    obj: v1.PromptVersionData,
    /,
    *,
    variables: Mapping[str, str] = MappingProxyType({}),
    formatter: Optional[TemplateFormatter] = None,
) -> tuple[list[protos.Content], GoogleModelKwargs]:
    formatter = formatter or to_formatter(obj)
    assert formatter is not None
    template = obj["template"]
    system_messages: list[str] = []
    messages: list[protos.Content] = []
    if template["type"] == "chat":
        for message in template["messages"]:
            if message["role"] == "system":
                for content in _ContentConversion.to_google(message, variables, formatter):
                    for part in content.parts:
                        if text := part.text:
                            system_messages.append(text)
            else:
                messages.extend(_ContentConversion.to_google(message, variables, formatter))
    elif template["type"] == "string":
        raise NotImplementedError
    else:
        assert_never(template)
    kwargs: GoogleModelKwargs = _to_model_kwargs(obj)
    if system_messages:
        if len(system_messages) == 1:
            kwargs["system_instruction"] = system_messages[0]
        else:
            kwargs["system_instruction"] = system_messages
    return messages, kwargs


def _to_model_kwargs(
    obj: v1.PromptVersionData,
    /,
) -> GoogleModelKwargs:
    invocation_parameters: v1.PromptGoogleInvocationParametersContent = (
        obj["invocation_parameters"]["google"]
        if "invocation_parameters" in obj and obj["invocation_parameters"]["type"] == "google"
        else {}
    )
    temperature = invocation_parameters.get("temperature")
    max_output_tokens = invocation_parameters.get("max_output_tokens")
    stop_sequences = invocation_parameters.get("stop_sequences")
    presence_penalty = invocation_parameters.get("presence_penalty")
    frequency_penalty = invocation_parameters.get("frequency_penalty")
    top_p = invocation_parameters.get("top_p")
    top_k = invocation_parameters.get("top_k")
    from google.generativeai.types import GenerationConfig

    generation_config = GenerationConfig(
        temperature=temperature,
        max_output_tokens=max_output_tokens,
        stop_sequences=stop_sequences,
        presence_penalty=presence_penalty,
        frequency_penalty=frequency_penalty,
        top_p=top_p,
        top_k=top_k,
    )
    return {
        "model_name": obj["model_name"],
        "generation_config": generation_config,
    }


class _ToolKwargsConversion:
    @staticmethod
    def to_google(
        obj: Optional[v1.PromptTools],
    ) -> _ToolKwargs:
        ans: _ToolKwargs = {}
        if not obj:
            return ans
        function_declarations: list[content_types.FunctionDeclaration] = []
        for t in obj["tools"]:
            if t["type"] == "function":
                function_declarations.append(_FunctionDeclarationConversion.to_google(t))
        from google.generativeai.types import content_types

        ans["tools"] = [
            content_types.Tool(
                function_declarations=function_declarations,
            )
        ]
        if "tool_choice" in obj:
            ans["tool_config"] = _ToolConfigConversion.to_google(obj["tool_choice"])
        return ans

    @staticmethod
    def from_google(
        obj: _ToolKwargs,
    ) -> Optional[v1.PromptTools]:
        if not obj:
            return None
        tools: list[v1.PromptToolFunction] = []
        if "tools" in obj:
            for tool in obj["tools"]:
                for fd in tool.function_declarations:
                    tools.append(_FunctionDeclarationConversion.from_google(fd))
        ans = v1.PromptTools(
            type="tools",
            tools=tools,
        )
        if "tool_config" in obj:
            ans["tool_choice"] = _ToolConfigConversion.from_google(obj["tool_config"])
        return ans


class _ToolConfigConversion:
    @staticmethod
    def to_google(
        obj: Union[
            v1.PromptToolChoiceNone,
            v1.PromptToolChoiceZeroOrMore,
            v1.PromptToolChoiceOneOrMore,
            v1.PromptToolChoiceSpecificFunctionTool,
        ],
    ) -> protos.ToolConfig:
        from google.generativeai import protos
        from google.generativeai.types import content_types

        ans: protos.ToolConfig = protos.ToolConfig()  # type: ignore[no-untyped-call]
        if obj["type"] == "none":
            ans.function_calling_config.mode = content_types.FunctionCallingMode.NONE
            return ans
        if obj["type"] == "zero_or_more":
            ans.function_calling_config.mode = content_types.FunctionCallingMode.AUTO
            return ans
        if obj["type"] == "one_or_more":
            ans.function_calling_config.mode = content_types.FunctionCallingMode.ANY
            return ans
        if obj["type"] == "specific_function":
            ans.function_calling_config.mode = content_types.FunctionCallingMode.ANY
            ans.function_calling_config.allowed_function_names = [obj["function_name"]]
            return ans
        else:
            assert_never(obj["type"])

    @staticmethod
    def from_google(
        obj: protos.ToolConfig,
    ) -> Union[
        v1.PromptToolChoiceNone,
        v1.PromptToolChoiceZeroOrMore,
        v1.PromptToolChoiceOneOrMore,
        v1.PromptToolChoiceSpecificFunctionTool,
    ]:
        from google.generativeai.types import content_types

        fcc: protos.FunctionCallingConfig = obj.function_calling_config
        if fcc.mode is content_types.FunctionCallingMode.NONE:
            choice_none: v1.PromptToolChoiceNone = {"type": "none"}
            return choice_none
        if fcc.mode is content_types.FunctionCallingMode.AUTO:
            choice_zero_or_more: v1.PromptToolChoiceZeroOrMore = {"type": "zero_or_more"}
            return choice_zero_or_more
        if fcc.mode is content_types.FunctionCallingMode.ANY:
            if not fcc.allowed_function_names:
                choice_one_or_more: v1.PromptToolChoiceOneOrMore = {"type": "one_or_more"}
                return choice_one_or_more
            choice_specific_function_tool: v1.PromptToolChoiceSpecificFunctionTool = {
                "type": "specific_function",
                "function_name": fcc.allowed_function_names[0],
            }
            return choice_specific_function_tool
        raise NotImplementedError


class _FunctionDeclarationConversion:
    @staticmethod
    def to_google(
        obj: v1.PromptToolFunction,
    ) -> content_types.FunctionDeclaration:
        from google.generativeai.types import content_types

        function = obj["function"]
        return content_types.FunctionDeclaration(
            name=function["name"],
            description=function["description"] if "description" in function else "",
            parameters=dict(function["parameters"]) if "parameters" in function else None,
        )

    @staticmethod
    def from_google(
        obj: Union[content_types.FunctionDeclaration, protos.FunctionDeclaration],
    ) -> v1.PromptToolFunction:
        return v1.PromptToolFunction(
            type="function",
            function=v1.PromptToolFunctionDefinition(
                name=obj.name,
                description=obj.description,
                parameters=_SchemaConversion.from_google(obj.parameters),
            ),
        )


class _SchemaConversion:
    @staticmethod
    def to_google(
        obj: Mapping[str, Any],
    ) -> protos.Schema:
        from google.generativeai import protos

        ans: protos.Schema = protos.Schema()  # type: ignore[no-untyped-call]
        if isinstance(type_ := obj.get("type"), str):
            if type_ == "string":
                ans.type_ = protos.Type.STRING
            elif type_ == "number":
                ans.type_ = protos.Type.NUMBER
            elif type_ == "integer":
                ans.type_ = protos.Type.INTEGER
            elif type_ == "boolean":
                ans.type_ = protos.Type.BOOLEAN
            elif type_ == "array":
                ans.type_ = protos.Type.ARRAY
            elif type_ == "object":
                ans.type_ = protos.Type.OBJECT
        if isinstance(format_ := obj.get("format"), str):
            ans.format_ = format_
        if isinstance(description := obj.get("description"), str):
            ans.description = description
        if isinstance(nullable := obj.get("nullable"), bool):
            ans.nullable = nullable
        if isinstance(enum := obj.get("enum"), Sequence):
            ans.enum = list(cast(Sequence[str], enum))
        if isinstance(items := obj.get("items"), Mapping):
            ans.items = _SchemaConversion.to_google(cast(Mapping[str, Any], items))
        if isinstance(max_items := obj.get("maxItems"), int):
            ans.max_items = max_items
        if isinstance(min_items := obj.get("minItems"), int):
            ans.min_items = min_items
        if isinstance(properties := obj.get("properties"), Mapping):
            ans.properties = {
                k: _SchemaConversion.to_google(v)
                for k, v in cast(Mapping[str, Mapping[str, Any]], properties).items()
            }
        if isinstance(required := obj.get("required"), Sequence):
            ans.required = list(cast(Sequence[str], required))
        return ans

    @staticmethod
    def from_google(
        obj: protos.Schema,
    ) -> dict[str, Any]:
        from google.generativeai import protos

        ans: dict[str, Any] = {}
        if obj.type_ is protos.Type.STRING:
            ans["type"] = "string"
        elif obj.type_ is protos.Type.NUMBER:
            ans["type"] = "number"
        elif obj.type_ is protos.Type.INTEGER:
            ans["type"] = "integer"
        elif obj.type_ is protos.Type.BOOLEAN:
            ans["type"] = "boolean"
        elif obj.type_ is protos.Type.ARRAY:
            ans["type"] = "array"
        elif obj.type_ is protos.Type.OBJECT:
            ans["type"] = "object"
        if obj.format_:
            ans["format"] = obj.format_
        if obj.description:
            ans["description"] = obj.description
        if obj.nullable:
            ans["nullable"] = obj.nullable
        if obj.enum:
            ans["enum"] = list(obj.enum)
        if obj.items:
            ans["items"] = _SchemaConversion.from_google(obj.items)
        if obj.max_items:
            ans["max_items"] = obj.max_items
        if obj.min_items:
            ans["min_items"] = obj.min_items
        if obj.properties:
            ans["properties"] = {
                k: _SchemaConversion.from_google(v) for k, v in obj.properties.items()
            }
        if obj.required:
            ans["required"] = list(obj.required)
        return ans


class _ContentConversion:
    @staticmethod
    def to_google(
        obj: v1.PromptMessage,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> Iterator[protos.Content]:
        from google.generativeai import protos

        role = _RoleConversion.to_google(obj)
        parts: list[protos.Part] = []
        if isinstance(obj["content"], str):
            text = formatter.format(obj["content"], variables=variables)
            yield protos.Content(role=role, parts=[protos.Part(text=text)])  # type: ignore[no-untyped-call]
            return
        for part in obj["content"]:
            if part["type"] == "text":
                parts.append(_TextContentPartConversion.to_google(part, variables, formatter))
            elif part["type"] == "tool_call":
                continue
            elif part["type"] == "tool_result":
                continue
            elif TYPE_CHECKING:
                assert_never(part["type"])
        yield protos.Content(role=role, parts=parts)  # type: ignore[no-untyped-call]

    @staticmethod
    def from_google(
        obj: protos.Content,
    ) -> v1.PromptMessage:
        role = _RoleConversion.from_google(obj)
        parts: list[_ContentPart] = []
        for part in obj.parts:
            if _has_text(part):
                parts.append(_TextContentPartConversion.from_google(part))
            elif _has_function_call(part):
                continue
            elif _has_function_response(part):
                continue
        return v1.PromptMessage(role=role, content=parts)


class _TextContentPartConversion:
    @staticmethod
    def to_google(
        obj: v1.TextContentPart,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> protos.Part:
        from google.generativeai import protos

        text = formatter.format(obj["text"], variables=variables)
        ans: protos.Part = protos.Part()  # type: ignore[no-untyped-call]
        ans.text = text
        return ans

    @staticmethod
    def from_google(
        obj: protos.Part,
    ) -> v1.TextContentPart:
        return v1.TextContentPart(
            type="text",
            text=obj.text,
        )


class _RoleConversion:
    @staticmethod
    def to_google(
        obj: v1.PromptMessage,
    ) -> Literal["user", "model"]:
        role = obj["role"]
        if role == "user":
            return "user"
        if role == "assistant":
            return "model"
        if role == "model":
            return "model"
        if role == "ai":
            return "model"
        if role == "system":
            raise NotImplementedError
        if role == "developer":
            raise NotImplementedError
        if role == "tool":
            return "user"
        if TYPE_CHECKING:
            assert_never(role)
        return role

    @staticmethod
    def from_google(
        obj: protos.Content,
    ) -> Literal["user", "assistant", "tool"]:
        if obj.role in ("model", "assistant"):
            return "assistant"
        if obj.role == "user":
            for part in obj.parts:
                p: protos.Part = part
                if _has_function_response(p):
                    return "tool"
                else:
                    continue
            return "user"
        return obj.role  # type: ignore


def _has_text(obj: protos.Part) -> bool:
    return bool(obj.text)


def _has_function_call(obj: protos.Part) -> bool:
    fc: protos.FunctionCall = obj.function_call
    return bool(fc.id or fc.name or fc.args)


def _has_function_response(obj: protos.Part) -> bool:
    fr: protos.FunctionResponse = obj.function_response
    return bool(fr.id or fr.name or fr.response)



from typing import TYPE_CHECKING, Any, Mapping, Optional, Union

if TYPE_CHECKING:
    from phoenix.client.client import AsyncClient, Client
from phoenix.client.resources.datasets import Dataset
from phoenix.client.resources.experiments.types import (
    ExperimentEvaluators,
    ExperimentTask,
    RanExperiment,
    RateLimitErrors,
)

DEFAULT_TIMEOUT_IN_SECONDS = 60


def run_experiment(
    *,
    dataset: Dataset,
    task: ExperimentTask,
    evaluators: Optional[ExperimentEvaluators] = None,
    experiment_name: Optional[str] = None,
    experiment_description: Optional[str] = None,
    experiment_metadata: Optional[Mapping[str, Any]] = None,
    rate_limit_errors: Optional[RateLimitErrors] = None,
    dry_run: Union[bool, int] = False,
    print_summary: bool = True,
    timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    client: Optional["Client"] = None,
) -> RanExperiment:
    """
    Run an experiment using a given dataset of examples.

    An experiment is a user-defined task that runs on each example in a dataset. The results from
    each experiment can be evaluated using any number of evaluators to measure the behavior of the
    task. The experiment and evaluation results are stored in the Phoenix database for comparison
    and analysis.

    A `task` is a synchronous function that returns a JSON serializable output. If the `task` is a
    function of one argument then that argument will be bound to the `input` field of the dataset
    example. Alternatively, the `task` can be a function of any combination of specific argument
    names that will be bound to special values:

    - `input`: The input field of the dataset example
    - `expected`: The expected or reference output of the dataset example
    - `reference`: An alias for `expected`
    - `metadata`: Metadata associated with the dataset example
    - `example`: The dataset `Example` object with all associated fields

    An `evaluator` is either a synchronous function that returns an evaluation result object,
    which can take any of the following forms:

    - an EvaluationResult dict with optional fields for score, label, explanation and metadata
    - a `bool`, which will be interpreted as a score of 0 or 1 plus a label of "True" or "False"
    - a `float`, which will be interpreted as a score
    - a `str`, which will be interpreted as a label
    - a 2-`tuple` of (`float`, `str`), which will be interpreted as (score, explanation)

    If the `evaluator` is a function of one argument then that argument will be bound to the
    `output` of the task. Alternatively, the `evaluator` can be a function of any combination
    of specific argument names that will be bound to special values:

    - `input`: The input field of the dataset example
    - `output`: The output of the task
    - `expected`: The expected or reference output of the dataset example
    - `reference`: An alias for `expected`
    - `metadata`: Metadata associated with the dataset example

    Phoenix also provides pre-built evaluators in the `phoenix.experiments.evaluators` module.

    Args:
        dataset: The dataset on which to run the experiment.
        task: The task to run on each example in the dataset.
        evaluators: A single evaluator or sequence of evaluators used to evaluate the results
            of the experiment. Defaults to None.
        experiment_name: The name of the experiment. Defaults to None.
        experiment_description: A description of the experiment. Defaults to None.
        experiment_metadata: Metadata to associate with the experiment. Defaults to None.
        rate_limit_errors: An exception or sequence of exceptions to adaptively throttle on.
            Defaults to None.
        dry_run: Run the experiment in dry-run mode. When set, experiment results will not be
            recorded in Phoenix. If True, the experiment will run on a random dataset example.
            If an integer, the experiment will run on a random sample of the dataset examples
            of the given size. Defaults to False.
        print_summary: Whether to print a summary of the experiment and evaluation results.
            Defaults to True.
        timeout: The timeout for the task execution in seconds. Use this to run longer tasks
            to avoid re-queuing the same task multiple times. Defaults to 60.
        client: A Phoenix client instance to use for the experiment. If not provided, a new client
            will be configured from environment variables. Defaults to None.

    Returns:
        A dictionary containing the experiment results.

    Raises:
        ValueError: If dataset format is invalid or has no examples.
        httpx.HTTPStatusError: If the API returns an error response.

    Example:
        Basic usage:
            >>> from phoenix.client.experiments import run_experiment
            >>> from phoenix.client import Client
            >>> client = Client()
            >>> dataset = client.datasets.get_dataset(dataset="my-dataset")
            >>>
            >>> def my_task(input):
            ...     return f"Hello {input['name']}"
            >>>
            >>> experiment = run_experiment(
            ...     dataset=dataset,
            ...     task=my_task,
            ...     experiment_name="greeting-experiment"
            ... )

        With client configuration:
            >>> experiment = run_experiment(
            ...     base_url="https://app.phoenix.arize.com",
            ...     api_key="your-api-key",
            ...     dataset=dataset,
            ...     task=my_task,
            ...     experiment_name="greeting-experiment"
            ... )

        With evaluators:
            >>> def accuracy_evaluator(output, expected):
            ...     return 1.0 if output == expected['text'] else 0.0
            >>>
            >>> experiment = run_experiment(
            ...     dataset=dataset,
            ...     task=my_task,
            ...     evaluators=[accuracy_evaluator],
            ...     experiment_name="evaluated-experiment"
            ... )

        Using dynamic binding for tasks:
            >>> def my_task(input, metadata, expected):
            ...     # Task can access multiple fields from the dataset example
            ...     context = metadata.get("context", "")
            ...     return f"Context: {context}, Input: {input}, Expected: {expected}"
            >>>
            >>> experiment = run_experiment(
            ...     dataset=dataset,
            ...     task=my_task,
            ...     experiment_name="dynamic-task"
            ... )

        Using dynamic binding for evaluators:
            >>> def my_evaluator(output, input, expected, metadata):
            ...     # Evaluator can access task output and example fields
            ...     score = calculate_similarity(output, expected)
            ...     return {"score": score, "label": "pass" if score > 0.8 else "fail"}
            >>>
            >>> experiment = run_experiment(
            ...     dataset=dataset,
            ...     task=my_task,
            ...     evaluators=[my_evaluator],
            ...     experiment_name="dynamic-evaluator"
            ... )

        Direct client usage (equivalent):
            >>> from phoenix.client import Client
            >>> client = Client()
            >>> experiment = client.experiments.run_experiment(
            ...     dataset=dataset,
            ...     task=my_task,
            ...     experiment_name="greeting-experiment"
            ... )
    """
    if client is None:
        from phoenix.client.client import Client

        client = Client()
    return client.experiments.run_experiment(
        dataset=dataset,
        task=task,
        evaluators=evaluators,
        experiment_name=experiment_name,
        experiment_description=experiment_description,
        experiment_metadata=experiment_metadata,
        rate_limit_errors=rate_limit_errors,
        dry_run=dry_run,
        print_summary=print_summary,
        timeout=timeout,
    )


async def async_run_experiment(
    *,
    dataset: Dataset,
    task: ExperimentTask,
    evaluators: Optional[ExperimentEvaluators] = None,
    experiment_name: Optional[str] = None,
    experiment_description: Optional[str] = None,
    experiment_metadata: Optional[Mapping[str, Any]] = None,
    rate_limit_errors: Optional[RateLimitErrors] = None,
    dry_run: Union[bool, int] = False,
    print_summary: bool = True,
    concurrency: int = 3,
    timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    client: Optional["AsyncClient"] = None,
) -> RanExperiment:
    """
    Run an experiment using a given dataset of examples (async version).

    An experiment is a user-defined task that runs on each example in a dataset. The results from
    each experiment can be evaluated using any number of evaluators to measure the behavior of the
    task. The experiment and evaluation results are stored in the Phoenix database for comparison
    and analysis.

    A `task` is either a synchronous or asynchronous function that returns a JSON serializable
    output. If the `task` is a function of one argument then that argument will be bound to the
    `input` field of the dataset example. Alternatively, the `task` can be a function of any
    combination of specific argument names that will be bound to special values:

    - `input`: The input field of the dataset example
    - `expected`: The expected or reference output of the dataset example
    - `reference`: An alias for `expected`
    - `metadata`: Metadata associated with the dataset example
    - `example`: The dataset `Example` object with all associated fields

    An `evaluator` is either a synchronous or asynchronous function that returns an evaluation
    result object, which can take any of the following forms:

    - an EvaluationResult dict with optional fields for score, label, explanation and metadata
    - a `bool`, which will be interpreted as a score of 0 or 1 plus a label of "True" or "False"
    - a `float`, which will be interpreted as a score
    - a `str`, which will be interpreted as a label
    - a 2-`tuple` of (`float`, `str`), which will be interpreted as (score, explanation)

    If the `evaluator` is a function of one argument then that argument will be bound to the
    `output` of the task. Alternatively, the `evaluator` can be a function of any combination
    of specific argument names that will be bound to special values:

    - `input`: The input field of the dataset example
    - `output`: The output of the task
    - `expected`: The expected or reference output of the dataset example
    - `reference`: An alias for `expected`
    - `metadata`: Metadata associated with the dataset example

    Phoenix also provides pre-built evaluators in the `phoenix.experiments.evaluators` module.

    Args:
        base_url: The base URL for the API endpoint. If not provided, it will be read from the
            environment variables or fall back to http://localhost:6006/.
        api_key: The API key for authentication. If provided, it will be included in the
            Authorization header as a bearer token. Defaults to None.
        headers: Additional headers to be included in the HTTP requests. Defaults to None.
            This is ignored if http_client is provided. Additional headers may be added from
            the environment variables, but won't override specified values.
        http_client: An instance of httpx.AsyncClient to be used for making HTTP requests. If not
            provided, a new instance will be created. Defaults to None.
        dataset: The dataset on which to run the experiment.
        task: The task to run on each example in the dataset.
        evaluators: A single evaluator or sequence of evaluators used to evaluate the results
            of the experiment. Defaults to None.
        experiment_name: The name of the experiment. Defaults to None.
        experiment_description: A description of the experiment. Defaults to None.
        experiment_metadata: Metadata to associate with the experiment. Defaults to None.
        rate_limit_errors: An exception or sequence of exceptions to adaptively throttle on.
            Defaults to None.
        dry_run: Run the experiment in dry-run mode. When set, experiment results will not be
            recorded in Phoenix. If True, the experiment will run on a random dataset example.
            If an integer, the experiment will run on a random sample of the dataset examples
            of the given size. Defaults to False.
        print_summary: Whether to print a summary of the experiment and evaluation results.
            Defaults to True.
        concurrency: Specifies the concurrency for task execution. Defaults to 3.
        timeout: The timeout for the task execution in seconds. Use this to run longer tasks
            to avoid re-queuing the same task multiple times. Defaults to 60.

    Returns:
        A dictionary containing the experiment results.

    Raises:
        ValueError: If dataset format is invalid or has no examples.
        httpx.HTTPStatusError: If the API returns an error response.

    Example:
        Basic usage:
            >>> from phoenix.client.experiments import async_run_experiment
            >>> from phoenix.client import AsyncClient
            >>> client = AsyncClient()
            >>> dataset = await client.datasets.get_dataset(dataset="my-dataset")
            >>>
            >>> async def my_task(input):
            ...     return f"Hello {input['name']}"
            >>>
            >>> experiment = await async_run_experiment(
            ...     dataset=dataset,
            ...     task=my_task,
            ...     experiment_name="greeting-experiment"
            ... )

        With client configuration:
            >>> experiment = await async_run_experiment(
            ...     base_url="https://app.phoenix.arize.com",
            ...     api_key="your-api-key",
            ...     dataset=dataset,
            ...     task=my_task,
            ...     experiment_name="greeting-experiment"
            ... )

        With evaluators:
            >>> async def accuracy_evaluator(output, expected):
            ...     return 1.0 if output == expected['text'] else 0.0
            >>>
            >>> experiment = await async_run_experiment(
            ...     dataset=dataset,
            ...     task=my_task,
            ...     evaluators=[accuracy_evaluator],
            ...     experiment_name="evaluated-experiment"
            ... )

        Using dynamic binding for tasks:
            >>> async def my_task(input, metadata, expected):
            ...     # Task can access multiple fields from the dataset example
            ...     context = metadata.get("context", "")
            ...     return f"Context: {context}, Input: {input}, Expected: {expected}"
            >>>
            >>> experiment = await async_run_experiment(
            ...     dataset=dataset,
            ...     task=my_task,
            ...     experiment_name="dynamic-task",
            ...     concurrency=5
            ... )

        Using dynamic binding for evaluators:
            >>> async def my_evaluator(output, input, expected, metadata):
            ...     # Evaluator can access task output and example fields
            ...     score = await calculate_similarity(output, expected)
            ...     return {"score": score, "label": "pass" if score > 0.8 else "fail"}
            >>>
            >>> experiment = await async_run_experiment(
            ...     dataset=dataset,
            ...     task=my_task,
            ...     evaluators=[my_evaluator],
            ...     experiment_name="dynamic-evaluator",
            ...     concurrency=10
            ... )

        Direct client usage (equivalent):
            >>> from phoenix.client import AsyncClient
            >>> client = AsyncClient()
            >>> experiment = await client.experiments.run_experiment(
            ...     dataset=dataset,
            ...     task=my_task,
            ...     experiment_name="greeting-experiment",
            ...     concurrency=5
            ... )
    """
    if client is None:
        from phoenix.client.client import AsyncClient

        client = AsyncClient()
    return await client.experiments.run_experiment(
        dataset=dataset,
        task=task,
        evaluators=evaluators,
        experiment_name=experiment_name,
        experiment_description=experiment_description,
        experiment_metadata=experiment_metadata,
        rate_limit_errors=rate_limit_errors,
        dry_run=dry_run,
        print_summary=print_summary,
        concurrency=concurrency,
        timeout=timeout,
    )


def get_experiment(
    *,
    experiment_id: str,
    client: Optional["Client"] = None,
) -> RanExperiment:
    """
    Get a completed experiment by ID.

    This function retrieves a completed experiment with all its task runs and evaluation runs,
    returning a RanExperiment object that can be used with evaluate_experiment to run additional
    evaluations.

    Args:
        experiment_id: The ID of the experiment to retrieve.
        client: A Phoenix client instance to use for the experiment. If not provided, a new client
            will be configured from environment variables. Defaults to None.

    Returns:
        A RanExperiment object containing the experiment data, task runs, and evaluation runs.

    Raises:
        ValueError: If the experiment is not found.
        httpx.HTTPStatusError: If the API returns an error response.

    Example:
        Basic usage:
            >>> from phoenix.client.experiments import get_experiment
            >>> experiment = get_experiment(experiment_id="123")

        Using with evaluate_experiment:
            >>> from phoenix.client.experiments import get_experiment, evaluate_experiment
            >>> experiment = get_experiment(experiment_id="123")
            >>> evaluated = evaluate_experiment(
            ...     experiment=experiment,
            ...     evaluators=[correctness_evaluator],
            ...     print_summary=True,
            ... )

        Direct client usage (equivalent):
            >>> from phoenix.client import Client
            >>> client = Client()
            >>> experiment = client.experiments.get_experiment(experiment_id="123")
    """
    if client is None:
        from phoenix.client.client import Client

        client = Client()
    return client.experiments.get_experiment(experiment_id=experiment_id)


async def async_get_experiment(
    *,
    experiment_id: str,
    client: Optional["AsyncClient"] = None,
) -> RanExperiment:
    """
    Get a completed experiment by ID (async version).

    This function retrieves a completed experiment with all its task runs and evaluation runs,
    returning a RanExperiment object that can be used with async_evaluate_experiment to run
    additional evaluations.

    Args:
        experiment_id: The ID of the experiment to retrieve.
        client: A Phoenix client instance to use for the experiment. If not provided, a new client
            will be configured from environment variables. Defaults to None.

    Returns:
        A RanExperiment object containing the experiment data, task runs, and evaluation runs.

    Raises:
        ValueError: If the experiment is not found.
        httpx.HTTPStatusError: If the API returns an error response.

    Example:
        Basic usage:
            >>> from phoenix.client.experiments import async_get_experiment
            >>> experiment = await async_get_experiment(experiment_id="123")

        Using with async_evaluate_experiment:
            >>> from phoenix.client.experiments import (
            ...     async_get_experiment,
            ...     async_evaluate_experiment,
            ... )
            >>> experiment = await async_get_experiment(experiment_id="123")
            >>> evaluated = await async_evaluate_experiment(
            ...     experiment=experiment,
            ...     evaluators=[correctness_evaluator],
            ...     print_summary=True,
            ... )

        Direct client usage (equivalent):
            >>> from phoenix.client import AsyncClient
            >>> client = AsyncClient()
            >>> experiment = await client.experiments.get_experiment(experiment_id="123")
    """
    if client is None:
        from phoenix.client.client import AsyncClient

        client = AsyncClient()
    return await client.experiments.get_experiment(experiment_id=experiment_id)


def evaluate_experiment(
    *,
    experiment: RanExperiment,
    evaluators: ExperimentEvaluators,
    dry_run: bool = False,
    print_summary: bool = True,
    timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    rate_limit_errors: Optional[RateLimitErrors] = None,
    client: Optional["Client"] = None,
) -> RanExperiment:
    """
    Run evaluators on a completed experiment.

    An evaluator is either a synchronous or asynchronous function that returns an evaluation
    result object, which can take any of the following forms:

    - an EvaluationResult dict with optional fields for score, label, explanation and metadata
    - a `bool`, which will be interpreted as a score of 0 or 1 plus a label of "True" or "False"
    - a `float`, which will be interpreted as a score
    - a `str`, which will be interpreted as a label
    - a 2-`tuple` of (`float`, `str`), which will be interpreted as (score, explanation)

    If the `evaluator` is a function of one argument then that argument will be bound to the
    `output` of the task. Alternatively, the `evaluator` can be a function of any combination
    of specific argument names that will be bound to special values:

    - `input`: The input field of the dataset example
    - `output`: The output of the task
    - `expected`: The expected or reference output of the dataset example
    - `reference`: An alias for `expected`
    - `metadata`: Metadata associated with the dataset example

    Phoenix also provides pre-built evaluators in the `phoenix.experiments.evaluators` module.

    Args:
        experiment: The experiment to evaluate, returned from `run_experiment` or `get_experiment`.
        evaluators: A single evaluator or sequence of evaluators used to evaluate the results
            of the experiment.
        dry_run: Run the evaluation in dry-run mode. When set, evaluation results will not be
            recorded in Phoenix. Defaults to False.
        print_summary: Whether to print a summary of the evaluation results. Defaults to True.
        timeout: The timeout for the evaluation execution in seconds. Defaults to 60.
        rate_limit_errors: An exception or sequence of exceptions to adaptively throttle on.
            Defaults to None.
        client: A Phoenix client instance to use for the experiment. If not provided, a new client
            will be configured from environment variables. Defaults to None.

    Returns:
        A dictionary containing the evaluation results with the same format as run_experiment.

    Raises:
        ValueError: If no evaluators are provided or experiment has no runs.
        httpx.HTTPStatusError: If the API returns an error response.

    Example:
        Basic usage:
            >>> from phoenix.client.experiments import get_experiment, evaluate_experiment
            >>> experiment = get_experiment(experiment_id="123")
            >>> def accuracy_evaluator(output, expected):
            ...     return 1.0 if output == expected else 0.0
            >>> evaluated = evaluate_experiment(
            ...     experiment=experiment,
            ...     evaluators=[accuracy_evaluator]
            ... )

        Using dynamic binding for evaluators:
            >>> def my_evaluator(output, input, expected, metadata):
            ...     # Evaluator can access task output and example fields
            ...     score = calculate_similarity(output, expected)
            ...     return {"score": score, "label": "pass" if score > 0.8 else "fail"}
            >>> evaluated = evaluate_experiment(
            ...     experiment=experiment,
            ...     evaluators=[my_evaluator]
            ... )

        Direct client usage (equivalent):
            >>> from phoenix.client import Client
            >>> client = Client()
            >>> evaluated = client.experiments.evaluate_experiment(
            ...     experiment=experiment,
            ...     evaluators=[accuracy_evaluator]
            ... )
    """
    if client is None:
        from phoenix.client.client import Client

        client = Client()
    return client.experiments.evaluate_experiment(
        experiment=experiment,
        evaluators=evaluators,
        dry_run=dry_run,
        print_summary=print_summary,
        timeout=timeout,
        rate_limit_errors=rate_limit_errors,
    )


async def async_evaluate_experiment(
    *,
    experiment: RanExperiment,
    evaluators: ExperimentEvaluators,
    dry_run: bool = False,
    print_summary: bool = True,
    timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    concurrency: int = 3,
    rate_limit_errors: Optional[RateLimitErrors] = None,
    client: Optional["AsyncClient"] = None,
) -> RanExperiment:
    """
    Run evaluators on a completed experiment (async version).

    An evaluator is either a synchronous or asynchronous function that returns an
    evaluation result object, which can take any of the following forms:

    - an EvaluationResult dict with optional fields for score, label, explanation and metadata
    - a `bool`, which will be interpreted as a score of 0 or 1 plus a label of "True" or "False"
    - a `float`, which will be interpreted as a score
    - a `str`, which will be interpreted as a label
    - a 2-`tuple` of (`float`, `str`), which will be interpreted as (score, explanation)

    If the `evaluator` is a function of one argument then that argument will be bound to the
    `output` of the task. Alternatively, the `evaluator` can be a function of any combination
    of specific argument names that will be bound to special values:

    - `input`: The input field of the dataset example
    - `output`: The output of the task
    - `expected`: The expected or reference output of the dataset example
    - `reference`: An alias for `expected`
    - `metadata`: Metadata associated with the dataset example

    Phoenix also provides pre-built evaluators in the `phoenix.experiments.evaluators` module.

    Args:
        experiment: The experiment to evaluate, returned from `async_run_experiment` or
            `async_get_experiment`.
        evaluators: A single evaluator or sequence of evaluators used to evaluate the results
            of the experiment.
        dry_run: Run the evaluation in dry-run mode. When set, evaluation results will not be
            recorded in Phoenix. Defaults to False.
        print_summary: Whether to print a summary of the evaluation results. Defaults to True.
        timeout: The timeout for the evaluation execution in seconds. Defaults to 60.
        concurrency: Specifies the concurrency for evaluation execution. Defaults to 3.
        rate_limit_errors: An exception or sequence of exceptions to adaptively throttle on.
            Defaults to None.
        client: A Phoenix client instance to use for the experiment. If not provided, a new client
            will be configured from environment variables. Defaults to None.

    Returns:
        A dictionary containing the evaluation results with the same format as async_run_experiment.

    Raises:
        ValueError: If no evaluators are provided or experiment has no runs.
        httpx.HTTPStatusError: If the API returns an error response.

    Example:
        Basic usage:
            >>> from phoenix.client.experiments import (
            ...     async_get_experiment,
            ...     async_evaluate_experiment,
            ... )
            >>> experiment = await async_get_experiment(experiment_id="123")
            >>> async def accuracy_evaluator(output, expected):
            ...     return 1.0 if output == expected else 0.0
            >>> evaluated = await async_evaluate_experiment(
            ...     experiment=experiment,
            ...     evaluators=[accuracy_evaluator]
            ... )

        Using dynamic binding for evaluators:
            >>> async def my_evaluator(output, input, expected, metadata):
            ...     # Evaluator can access task output and example fields
            ...     score = await calculate_similarity(output, expected)
            ...     return {"score": score, "label": "pass" if score > 0.8 else "fail"}
            >>> evaluated = await async_evaluate_experiment(
            ...     experiment=experiment,
            ...     evaluators=[my_evaluator],
            ...     concurrency=10
            ... )

        Direct client usage (equivalent):
            >>> from phoenix.client import AsyncClient
            >>> client = AsyncClient()
            >>> evaluated = await client.experiments.evaluate_experiment(
            ...     experiment=experiment,
            ...     evaluators=[accuracy_evaluator],
            ...     concurrency=5
            ... )
    """
    if client is None:
        from phoenix.client.client import AsyncClient

        client = AsyncClient()
    return await client.experiments.evaluate_experiment(
        experiment=experiment,
        evaluators=evaluators,
        dry_run=dry_run,
        print_summary=print_summary,
        timeout=timeout,
        concurrency=concurrency,
        rate_limit_errors=rate_limit_errors,
    )
